"""Group2 Cookbook2.

Multiline doc that will be cut to one-line in the __title__.
"""
import argparse

__title__ = __doc__


def argument_parser():
    """As required by spicerack._cookbook."""
    parser = argparse.ArgumentParser("Argparse")
    parser.add_argument("-k", action="store_true")
    parser.add_argument("--argument")

    return parser


def run(args, spicerack):
    """As required by spicerack._cookbook."""
    print([dict(sorted(vars(args).items())), spicerack.verbose, spicerack.dry_run])
    return 0



"""Call another cookbook cookbook."""
import argparse

from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class CallAnotherCookbook(CookbookBase):
    """A cookbook that calls another cookbook."""

    def argument_parser(self):
        """As defined by the parent class."""
        parser = argparse.ArgumentParser("Argparse")
        parser.add_argument("cookbook", help="The path of the cookbook to execute.")
        return parser

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner(args, self.spicerack)


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def __init__(self, args, spicerack):
        """Initialize the instance."""
        self.cookbook = args.cookbook
        self.spicerack = spicerack

    def run(self):
        """As required by the parent class."""
        return self.spicerack.run_cookbook(self.cookbook, [])



"""Puppet module tests."""
import json
from datetime import datetime, timedelta, timezone
from subprocess import CalledProcessError
from unittest import mock

import pytest
from ClusterShell.MsgTree import MsgTreeElem
from cumin import nodeset

from spicerack import puppet
from spicerack.administrative import Reason
from spicerack.remote import RemoteExecutionError, RemoteHosts

REASON = Reason("Disable reason", "user1", "orchestration-host", task_id="T12345")
PUPPET_CA_CERT_METADATA_SIGNED = (
    b'[{"name":"test.example.com","state":"signed","fingerprint":"00:AA",'
    b'"fingerprints":{"default":"00:AA","SHA1":"11:BB","SHA256": "00:AA","SHA512":"22:CC"},'
    b'"dns_alt_names":["DNS:service.example.com"]}]'
)
PUPPET_CA_CERT_METADATA_REQUESTED = (
    b'[{"name":"test.example.com","state":"requested","fingerprint":"00:AA",'
    b'"fingerprints":{"default":"00:AA","SHA1":"11:BB","SHA256": "00:AA","SHA512":"22:CC"},'
    b'"dns_alt_names":["DNS:service.example.com"]}]'
)
PUPPET_GENERATE_CERTIFICATE_SUCCESS = b"""Info: Creating a new SSL key for test.example.com
Info: Caching certificate for ca
Info: csr_attributes file loading from /etc/puppet/csr_attributes.yaml
Info: Creating a new SSL certificate request for test.example.com
Info: Certificate Request fingerprint (SHA256): 00:FF
Info: Caching certificate for ca
Exiting; no certificate found and waitforcert is disabled
"""
PUPPET_GENERATE_CERTIFICATE_FAILED = rb"""Info: Creating a new SSL key for test.example.com
Info: Caching certificate for ca
Info: Caching certificate for test.example.com
Error: Could not request certificate: The certificate retrieved from the master does not match the agent's private key.
Certificate fingerprint: 00:FF
To fix this, remove the certificate from both the master and the agent and then start a puppet run, which will...
On the master:
  puppet cert clean test.example.com
On the agent:
  1a. On most platforms: find /var/lib/puppet/ssl -name test.example.com.pem -delete
  1b. On Windows: del "\var\lib\puppet\ssl\certs\test.example.com.pem" /f
  2. puppet agent -t
Exiting; failed to retrieve certificate and waitforcert is disabled
"""


@mock.patch("spicerack.puppet.check_output", return_value=b"puppetmaster.example.com")
def test_get_puppet_ca_hostname_ok(mocked_check_output):
    """It should get the hostname of the Puppet CA from the local Puppet agent."""
    ca = puppet.get_puppet_ca_hostname()
    assert ca == "puppetmaster.example.com"
    mocked_check_output.assert_called_once_with(["puppet", "config", "print", "--section", "agent", "ca_server"])


@mock.patch(
    "spicerack.puppet.check_output",
    side_effect=CalledProcessError(1, "executed_command"),
)
def test_get_puppet_ca_hostname_fail(mocked_check_output):
    """It should raise PuppetMasterError if unable to get the hostname of the Puppet CA from the Puppet agent."""
    with pytest.raises(puppet.PuppetMasterError, match="Get Puppet ca_server failed"):
        puppet.get_puppet_ca_hostname()

    assert mocked_check_output.called


@mock.patch("spicerack.puppet.check_output", return_value=b"")
def test_get_puppet_ca_hostname_empty(mocked_check_output):
    """It should raise PuppetMasterError if the Puppet CA from Puppet is empty."""
    with pytest.raises(puppet.PuppetMasterError, match="Got empty ca_server from Puppet agent"):
        puppet.get_puppet_ca_hostname()

    assert mocked_check_output.called


class TestPuppetHosts:
    """Test class for the PuppetHosts class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_remote_hosts = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_remote_hosts.__len__.return_value = 1
        self.puppet_hosts = puppet.PuppetHosts(self.mocked_remote_hosts)

    @pytest.mark.parametrize(
        "verbatim_reason, result",
        ((False, f"disable-puppet {REASON.quoted()}"), (True, f'disable-puppet "{REASON.reason}"')),
    )
    def test_disable(self, verbatim_reason, result):
        """It should disable Puppet on the hosts."""
        self.puppet_hosts.disable(REASON, verbatim_reason)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(result)

    @pytest.mark.parametrize(
        "verbatim_reason, result",
        ((False, f"enable-puppet {REASON.quoted()}"), (True, f'enable-puppet "{REASON.reason}"')),
    )
    def test_enable(self, verbatim_reason, result):
        """It should enable Puppet on the hosts."""
        self.puppet_hosts.enable(REASON, verbatim_reason)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(result)

    @pytest.mark.parametrize("verbatim_reason, result", ((False, REASON.quoted()), (True, f'"{REASON.reason}"')))
    def test_disabled(self, verbatim_reason, result):
        """It should disable Puppet, yield and enable Puppet on the hosts."""
        with self.puppet_hosts.disabled(REASON, verbatim_reason):
            self.mocked_remote_hosts.run_sync.assert_called_once_with(f"disable-puppet {result}")
            self.mocked_remote_hosts.run_sync.reset_mock()

        self.mocked_remote_hosts.run_sync.assert_called_once_with(f"enable-puppet {result}")

    def test_disabled_on_raise(self):
        """It should re-enable Puppet even if the yielded code raises exception.."""
        with pytest.raises(RuntimeError):
            with self.puppet_hosts.disabled(REASON):
                self.mocked_remote_hosts.run_sync.reset_mock()
                raise RuntimeError("Error")

        self.mocked_remote_hosts.run_sync.assert_called_once_with(f"enable-puppet {REASON.quoted()}")

    def test_check_enabled_ok(self):
        """It should check that all hosts have Puppet enabled."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"0", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"0", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.puppet_hosts.check_enabled()

    def test_check_enabled_raise(self):
        """It should raise PuppetHostsCheckError if Puppet is disabled on some hosts."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"0", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"1", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        with pytest.raises(
            puppet.PuppetHostsCheckError,
            match="Puppet is not enabled on those hosts: test2.example.com",
        ):
            self.puppet_hosts.check_enabled()

    def test_check_disabled_ok(self):
        """It should check that all hosts have Puppet disabled."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"1", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"1", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.puppet_hosts.check_disabled()

    def test_check_disabled_raise(self):
        """It should raise PuppetHostsCheckError if Puppet is disabled on some hosts."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"1", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"0", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        with pytest.raises(
            puppet.PuppetHostsCheckError,
            match="Puppet is not disabled on those hosts: test2.example.com",
        ):
            self.puppet_hosts.check_disabled()

    @pytest.mark.parametrize(
        "kwargs, expected",
        (
            ({}, ""),
            ({"enable_reason": REASON}, "--enable " + REASON.quoted()),
            ({"quiet": True}, "--quiet"),
            ({"failed_only": True}, "--failed-only"),
            ({"force": True}, "--force"),
            ({"attempts": 5}, "--attempts 5"),
        ),
    )
    def test_run_ok(self, kwargs, expected):
        """It should run Puppet with the specified arguments."""
        self.puppet_hosts.run(**kwargs)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            puppet.Command(f"run-puppet-agent {expected}", timeout=300.0),
            batch_size=10,
        )

    def test_run_timeout(self):
        """It should run Puppet with the customized timeout."""
        self.puppet_hosts.run(timeout=30)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            puppet.Command("run-puppet-agent ", timeout=30.0), batch_size=10
        )

    def test_first_run(self):
        """It should enable and Puppet with a very long timeout without using custom wrappers."""
        self.puppet_hosts.first_run()
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            "systemctl stop puppet.service",
            "systemctl reset-failed puppet.service || true",
            "puppet agent --enable",
            puppet.Command(
                (
                    "puppet agent --onetime --no-daemonize --verbose --no-splay --show_diff --ignorecache "
                    "--no-usecacheonfailure"
                ),
                timeout=10800,
            ),
            print_output=False,
            print_progress_bars=False,
        )

    def test_first_run_not_systemd(self):
        """It should enable and Puppet with a very long timeout without using custom wrappers."""
        self.puppet_hosts.first_run(has_systemd=False)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            "puppet agent --enable",
            puppet.Command(
                (
                    "puppet agent --onetime --no-daemonize --verbose --no-splay --show_diff --ignorecache "
                    "--no-usecacheonfailure"
                ),
                timeout=10800,
            ),
            print_output=False,
            print_progress_bars=False,
        )

    def test_regenerate_certificate_ok(self):
        """It should delete and regenerate the Puppet certificate."""
        results = [
            (
                nodeset("test.example.com"),
                MsgTreeElem(PUPPET_GENERATE_CERTIFICATE_SUCCESS, parent=MsgTreeElem()),
            )
        ]
        self.mocked_remote_hosts.run_sync.side_effect = [iter(()), iter(results)]

        fingerprints = self.puppet_hosts.regenerate_certificate()

        self.mocked_remote_hosts.run_sync.assert_has_calls(
            [
                mock.call("rm -rfv /var/lib/puppet/ssl"),
                mock.call(puppet.Command("puppet agent --test --color=false", ok_codes=[]), print_output=False),
            ]
        )
        assert fingerprints == {"test.example.com": "00:FF"}

    def test_regenerate_certificate_raise(self):
        """It should raise PuppetHostsError if unable to find any of the fingerprint."""
        message = PUPPET_GENERATE_CERTIFICATE_SUCCESS.decode().replace(": 00:FF", "").encode()
        results = [(nodeset("test.example.com"), MsgTreeElem(message, parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.side_effect = [iter(()), iter(results)]

        with pytest.raises(
            puppet.PuppetHostsError,
            match="Unable to find CSR fingerprints for all hosts",
        ):
            self.puppet_hosts.regenerate_certificate()

    def test_regenerate_certificate_errors(self):
        """It should raise PuppetHostsError and print the Puppet errors if unable to find any of the fingerprint."""
        results = [
            (
                nodeset("test.example.com"),
                MsgTreeElem(PUPPET_GENERATE_CERTIFICATE_FAILED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_remote_hosts.run_sync.side_effect = [iter(()), iter(results)]

        with pytest.raises(
            puppet.PuppetHostsError,
            match=(
                "test.example.com: Error: Could not request certificate: The certificate retrieved "
                "from the master does not match the agent's private key."
            ),
        ):
            self.puppet_hosts.regenerate_certificate()

    def test_wait_since_ok(self):
        """It should return immediately if there is already successful Puppet run since the given datetime."""
        last_run = datetime.utcnow()
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))
        start = last_run - timedelta(seconds=1)

        nodes = nodeset("test.example.com")
        results = [(nodes, MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.mocked_remote_hosts.hosts = nodes

        self.puppet_hosts.wait_since(start)

        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            (
                "source /usr/local/share/bash/puppet-common.sh && last_run_success && awk /last_run/'{ print $2 }' "
                '"${PUPPET_SUMMARY}"'
            ),
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    # TODO: check why the following test_wait_since_* tests take longer (~4s each) when running the whole suite but are
    # quick if running only tests in this module (tox -e py34-unit -- -k test_puppet)
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_since_timeout(self, mocked_sleep):
        """It should raise PuppetHostsCheckError if the successful Puppet run is too old within the timeout."""
        last_run = datetime.utcnow()
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))
        start = last_run + timedelta(seconds=1)

        nodes = nodeset("test.example.com")
        results = [(nodes, MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.side_effect = [results] * 60
        self.mocked_remote_hosts.hosts = nodes

        with pytest.raises(puppet.PuppetHostsCheckError, match="Successful Puppet run too old"):
            self.puppet_hosts.wait_since(start)

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_since_failed_execution(self, mocked_sleep):
        """It should raise PuppetHostsCheckError if fails to get the successful Puppet run within the timeout."""
        self.mocked_remote_hosts.run_sync.side_effect = RemoteExecutionError(1, "fail", iter(()))
        self.mocked_remote_hosts.hosts = nodeset("test.example.com")

        with pytest.raises(puppet.PuppetHostsCheckError, match="Unable to find a successful Puppet run"):
            self.puppet_hosts.wait_since(datetime.utcnow())

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_since_missing_host(self, mocked_sleep):
        """It should raise PuppetHostsCheckError unable to get the result from some host."""
        last_run = datetime.utcnow()
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))
        start = last_run - timedelta(seconds=1)

        nodes = nodeset("test[1-2].example.com")
        results = [
            (
                nodeset("test1.example.com"),
                MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()),
            )
        ]
        self.mocked_remote_hosts.run_sync.side_effect = [results] * 60
        self.mocked_remote_hosts.hosts = nodes

        with pytest.raises(
            puppet.PuppetHostsCheckError,
            match="Unable to get successful Puppet run from: test2.example.com",
        ):
            self.puppet_hosts.wait_since(start)

        assert mocked_sleep.called

    def test_wait_ok(self):
        """It should return immediately if there is already successful Puppet run since now."""
        last_run = datetime.utcnow() + timedelta(seconds=1)
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))

        nodes = nodeset("test.example.com")
        results = [(nodes, MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.mocked_remote_hosts.hosts = nodes

        self.puppet_hosts.wait()

        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            (
                "source /usr/local/share/bash/puppet-common.sh && last_run_success && awk /last_run/'{ print $2 }' "
                '"${PUPPET_SUMMARY}"'
            ),
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_get_ca_servers_explodes_multihost_nodeset_into_single_hosts(self):
        """Test that get ca servers explodes multihost nodeset into single hosts."""
        expected_puppetmaster = "dummy.puppetmast.er"
        self.mocked_remote_hosts.run_sync.return_value = [
            (
                nodeset("test[0-1].example.com"),
                MsgTreeElem(expected_puppetmaster.encode(), parent=MsgTreeElem()),
            ),
        ]

        result = self.puppet_hosts.get_ca_servers()

        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert "test0.example.com" in result
        assert result["test0.example.com"] == expected_puppetmaster
        assert "test1.example.com" in result
        assert result["test1.example.com"] == expected_puppetmaster

    def test_get_ca_servers_handles_empty_result(self):
        """Test that get ca servers handles empty result."""
        self.mocked_remote_hosts.run_sync.return_value = []

        result = self.puppet_hosts.get_ca_servers()

        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert result == {}  # pylint: disable=use-implicit-booleaness-not-comparison

    def test_get_ca_servers_handles_multiple_results(self):
        """Test test get ca servers handles multiple results."""
        self.mocked_remote_hosts.run_sync.return_value = [
            (
                nodeset("test0.example.com"),
                MsgTreeElem(b"test0.puppetmast.er", parent=MsgTreeElem()),
            ),
            (
                nodeset("test1.example.com"),
                MsgTreeElem(b"test1.puppetmast.er", parent=MsgTreeElem()),
            ),
        ]

        result = self.puppet_hosts.get_ca_servers()

        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert "test0.example.com" in result
        assert result["test0.example.com"] == "test0.puppetmast.er"
        assert "test1.example.com" in result
        assert result["test1.example.com"] == "test1.puppetmast.er"

    def test_get_ca_servers_handles_multiple_lines_in_command_output(self):
        """Test test get ca servers handles multiple results."""
        self.mocked_remote_hosts.run_sync.return_value = [
            (
                nodeset("test0.example.com"),
                MsgTreeElem(
                    b"test0.puppetmast.er",
                    parent=MsgTreeElem(b"some message that should be ignored", parent=MsgTreeElem()),
                ),
            ),
        ]

        result = self.puppet_hosts.get_ca_servers()

        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert "test0.example.com" in result
        assert result["test0.example.com"] == "test0.puppetmast.er"


class TestPuppetMaster:
    """Test class for the PuppetMaster class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_master_host = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_master_host.__len__.return_value = 1
        self.puppet_master = puppet.PuppetMaster(self.mocked_master_host)

    def test_init_raise(self):
        """It should raise PuppetMasterError if the Puppet master host instance doesn't match one host."""
        self.mocked_master_host.__len__.return_value = 2
        with pytest.raises(
            puppet.PuppetMasterError,
            match="The master_host instance must target only one host, got 2",
        ):
            puppet.PuppetMaster(self.mocked_master_host)

    def test_master_host(self):
        """It should return the master host RemoteHosts instance."""
        assert self.puppet_master.master_host is self.mocked_master_host

    def test_delete(self):
        """It should delete the host from Puppet master and PuppetDB."""
        self.puppet_master.delete("test.example.com")
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet node clean test.example.com", "puppet node deactivate test.example.com", print_progress_bars=False
        )

    def test_destroy(self):
        """It should delete the certificate of the host in the Puppet CA."""
        self.puppet_master.destroy("test.example.com")
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations destroy test.example.com", print_progress_bars=False
        )

    def test_verify_ok(self):
        """It should verify that the host has a signed certificate in the Puppet CA."""
        json_output = b'{"host":"test.example.com","valid":true}'
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        self.puppet_master.verify("test.example.com")
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json verify test.example.com",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_verify_raise(self):
        """It should raise PuppetMasterError if the certificate is not valid on the Puppet CA."""
        json_output = b'{"host":"test.example.com","valid":false,"error":"Error message"}'
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetMasterError,
            match="Invalid certificate for test.example.com: Error message",
        ):
            self.puppet_master.verify("test.example.com")

        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json verify test.example.com",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_verify_no_output(self):
        """It should raise PuppetMasterError if there is no output from the executed command."""
        self.mocked_master_host.run_sync.return_value = iter(())
        with pytest.raises(
            puppet.PuppetMasterError,
            match=(
                "Got no output from Puppet master while executing command: "
                "puppet ca --disable_warnings deprecations --render-as json verify test.example.com"
            ),
        ):
            self.puppet_master.verify("test.example.com")

    def test_verify_invalid_json(self):
        """It should raise PuppetMasterError if there output of the executed command is invalid JSON."""
        json_output = b'{"host":"test.example.com",,}'
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetMasterError,
            match=(
                "Unable to parse Puppet master response for command "
                '"puppet ca --disable_warnings deprecations --render-as json verify test.example.com"'
            ),
        ):
            self.puppet_master.verify("test.example.com")

    def test_sign_ok(self):
        """It should sign the certificate in the Puppet CA."""
        requested_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.side_effect = [
            iter(requested_results),
            iter(()),
            iter(signed_results),
        ]

        self.puppet_master.sign("test.example.com", "00:AA")

        self.mocked_master_host.run_sync.assert_has_calls(
            [
                mock.call(
                    "puppet ca --disable_warnings deprecations --render-as json "
                    r'list --all --subject "^test\.example\.com$"',
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppet cert --disable_warnings deprecations sign --no-allow-dns-alt-names test.example.com",
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppet ca --disable_warnings deprecations --render-as json "
                    r'list --all --subject "^test\.example\.com$"',
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
            ]
        )

    def test_sign_alt_dns(self):
        """It should pass the --allow-dns-alt-names option while signing the certificate."""
        requested_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.side_effect = [
            iter(requested_results),
            iter(()),
            iter(signed_results),
        ]

        self.puppet_master.sign("test.example.com", "00:AA", allow_alt_names=True)

        self.mocked_master_host.run_sync.assert_has_calls(
            [
                mock.call(
                    "puppet ca --disable_warnings deprecations --render-as json "
                    r'list --all --subject "^test\.example\.com$"',
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppet cert --disable_warnings deprecations sign --allow-dns-alt-names test.example.com",
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppet ca --disable_warnings deprecations --render-as json "
                    r'list --all --subject "^test\.example\.com$"',
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
            ]
        )

    def test_sign_wrong_state(self):
        """It should raise PuppetMasterError if the certificate is not in the requested state."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetMasterError,
            match="Certificate for test.example.com not in requested state, got: signed",
        ):
            self.puppet_master.sign("test.example.com", "00:AA")

    def test_sign_wrong_fingerprint(self):
        """It should raise PuppetMasterError if the fingerprint doesn't match the one in the CSR."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetMasterError,
            match="CSR fingerprint 00:AA for test.example.com does not match provided fingerprint FF:FF",
        ):
            self.puppet_master.sign("test.example.com", "FF:FF")

    def test_sign_fail(self):
        """It should raise PuppetMasterError if the sign operation fails."""
        requested_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        sign_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(b"sign error", parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.side_effect = [
            iter(requested_results),
            iter(sign_results),
            iter(signed_results),
        ]

        with pytest.raises(
            puppet.PuppetMasterError,
            match="Expected certificate for test.example.com to be signed, got: requested",
        ):
            self.puppet_master.sign("test.example.com", "00:AA")

    def test_wait_for_csr_already_ok(self):
        """It should return immediately if the certificate is already requested."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        self.puppet_master.wait_for_csr("test.example.com")

        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json "
            r'list --all --subject "^test\.example\.com$"',
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_wait_for_csr_fail(self):
        """It should raise PuppetMasterError if the certificate is in a wrong state."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetMasterError,
            match="Expected certificate in requested state, got: signed",
        ):
            self.puppet_master.wait_for_csr("test.example.com")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_csr_timeout(self, mocked_sleep):
        """It should raise PuppetMasterCheckError if the certificate request doesn't appear."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(b"[]", parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.side_effect = [iter(results) for _ in range(10)]

        with pytest.raises(
            puppet.PuppetMasterCheckError,
            match="No certificate found for hostname: test.example.com",
        ):
            self.puppet_master.wait_for_csr("test.example.com")

        assert mocked_sleep.called

    def test_get_certificate_metadata_ok(self):
        """It should return the metadata of the certificate for the host in the Puppet CA."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        metadata = self.puppet_master.get_certificate_metadata("test.example.com")

        assert metadata == json.loads(PUPPET_CA_CERT_METADATA_SIGNED.decode())[0]
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json list "
            r'--all --subject "^test\.example\.com$"',
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    @pytest.mark.parametrize(
        "json_output, exception_message",
        (
            (
                b'[{"name":"test.example.com"},{"name":"test.example.com"}]',
                "Expected one result from Puppet CA, got 2",
            ),
            (
                b'[{"name":"invalid.example.com"}]',
                "Hostname mismatch invalid.example.com != test.example.com",
            ),
        ),
    )
    def test_get_certificate_metadata_raises(self, json_output, exception_message):
        """It should raise PuppetMasterError if the Puppet CA returns multiple certificates metadata."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(puppet.PuppetMasterError, match=exception_message):
            self.puppet_master.get_certificate_metadata("test.example.com")



"""Interactive module."""
import logging
import os

from wmflib.interactive import get_secret

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


def get_management_password() -> str:
    """Get the management password either from the environment or asking for it.

    Raises:
        spicerack.exceptions.SpicerackError: if the password is empty.

    """
    password = os.getenv("MGMT_PASSWORD")

    if password is None:
        logger.debug("MGMT_PASSWORD environment variable not found")
        # Ask for a password, raise exception if not a tty
        password = get_secret("Management Password")
    else:
        logger.info("Using Management Password from the MGMT_PASSWORD environment variable")

    if not password:
        raise SpicerackError("Empty Management Password")

    return password



"""Class API rollback cookbook."""
import argparse
import logging

from spicerack.cookbook import CookbookBase, CookbookRunnerBase

logger = logging.getLogger()


class RollbackCookbook(CookbookBase):
    """Class API rollback cookbook."""

    def argument_parser(self):
        """As defined by the parent class."""
        return argparse.ArgumentParser("Argparse")

    def get_runner(self, args):
        """As required by the parent class."""
        return RollbackRunner()


class RollbackRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def run(self):
        """As required by the parent class."""
        raise RuntimeError("run has raised")

    def rollback(self):
        """Define rollback actions."""
        logger.error("rollback called")



"""Cookbook internal module."""
import argparse
import logging
import shlex
import sys
from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import Any, Optional, cast

from spicerack import Spicerack, _log, _module_api, cookbook
from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)
HELP_MESSAGE = """Cookbooks interactive menu help

Available cookbooks and cookbook groups are shown in the menu with the format:
  [STATUS] NAME: DESCRIPTION
Additional control commands are also shown.
To select an item just input its name and press Enter.

Group of cookbooks:
  They have a status that represent the number of the executed cookbooks over
  the total number of cookbooks in that group and its child groups
  (i.e. [2/11]) or the status 'DONE' in case all cookbooks in that group were
  executed during the current session.
  When selected the child cookbooks group is shown.

Single cookbooks:
  Their status has one of the following values:
  {statuses}
  When selected the cookbook is executed and then the current menu is shown
  again after its execution, with the status updated based on the result of the
  execution.

Control commands:
  b: shown when inside a child group of cookbooks to go back one level to the
     parent menu.
  q: shown when at the top level of the current session menu to exit the
     program.
  h: always shown, print this help message.

  Note: 'q' and 'b' are mutually exclusive, only one of them is shown.

CLI arguments:
  It's possible to pass CLI arguments to cookbooks and group of cookbooks
  when selecting them (e.g. cookbook_name -a param value1 value2).
  Passing arguments to cookbook groups propagate them also to their cookbooks
  and child cookbook groups.
  Passing arguments override any other argument that might have been passed
  to the cookbook executable or to any of the parent groups when selected.

Interrupting execution:
  Pressing Ctrl+c/d while executing a cookbook interrupts it and show the
  current menu, marking the cookbook status as ERROR.
  Pressing Ctrl+c/d while in a menu is equivalent to select 'b' or 'q'.
"""
"""The generic TreeItem help message, unformatted."""


class MenuError(SpicerackError):
    """Custom exception class for errors of this module."""


class BaseItem:
    """Base class for any item collected by the CookbookCollection class."""

    fallback_title: str = "-"

    def __init__(self, args: Sequence[str], spicerack: Spicerack) -> None:
        """Base cookbooks's item constructor.

        Arguments:
            args: any sequence with the command line arguments to pass to the item.
            spicerack: the initialized instance of the library.

        """
        self.args = args
        self.spicerack = spicerack
        self.name = ""
        self.path = ""  # Path of the cookbook in Spicerack terms, relative to the base directory
        self.full_name = ""  # Cookbook full_name in Spicerack terms
        self.items: dict[str, BaseItem] = {}
        self._status: str

    @abstractmethod
    def run(self) -> Optional[int]:
        """Excecute the item."""

    @property
    def status(self) -> str:
        """Return the current execution status string representation."""
        return self._status

    @property
    def verbose_title(self) -> str:
        """Getter for the verbose_title property, uses the module name if there is no title set."""
        if self.title != self.fallback_title:
            return self.title

        return self.name

    @property
    @abstractmethod
    def title(self) -> str:
        """Return the title of the instance item."""

    @staticmethod
    def _get_line_prefix(level: int, cont_levels: list[bool], is_final: bool) -> str:
        """Return the line prefix for the given level in the tree.

        Arguments:
            level: how many levels the item is nested in the tree.
            cont_levels: a list of size levels with booleans that indicate for each level if the continuation prefix
                (:py:data:`True`) or an empty prefix (:py:data:`False`) should be used.
            is_final: whether the line is the final in its own group.

        """
        empty_sep = "    "
        cont_sep = "|   "

        if is_final:
            base_sep = "`-- "
        else:
            base_sep = "|-- "

        if level == 0:
            return base_sep

        levels = []
        for cont_level in cont_levels:
            if cont_level:
                levels.append(cont_sep)
            else:
                levels.append(empty_sep)

        return "".join(levels) + base_sep


class CookbookItem(BaseItem):
    """Cookbook item class."""

    fallback_title: str = "UNKNOWN (unable to detect title)"
    statuses: tuple[str, str, str, str] = (
        "NOTRUN",
        "PASS",
        "FAIL",
        "ERROR",
    )  # Status labels
    not_run, success, failed, error = statuses  # Valid statuses variables

    def __init__(
        self,
        class_obj: type[cookbook.CookbookBase],
        args: Sequence[str],
        spicerack: Spicerack,
    ) -> None:
        """Override parent constructor to add cookbook-specific initialization.

        Arguments:
            class_obj: a class derived from :py:class:`spicerack.cookbook.CookbookBase`.
            args: any sequence with the command line arguments to pass to the item.
            spicerack: the initialized instance of the library.

        """
        super().__init__(args, spicerack)
        if not issubclass(class_obj, cookbook.CookbookBase):
            raise MenuError(f"Class {class_obj} is not a subclass of CookbookBase")

        self._status = CookbookItem.not_run
        self.name = class_obj.spicerack_name
        self.path = class_obj.spicerack_path
        if self.path:
            self.full_name = ".".join([self.path, self.name])
        else:
            self.full_name = self.name

        self.instance = class_obj(spicerack)

    @property
    def title(self) -> str:
        """Returns the title of the instance item."""
        return self.instance.title

    def run(self) -> int:  # noqa: MC0001
        """Run the cookbook.

        Returns:
            The return code to use for this cookbook, it should be zero on success, a positive integer smaller than
            ``128`` and not in the range ``90-99`` (see :ref:`Reserved exit codes<reserved-codes>`) in case of failure.

        """
        ret, args = self._parse_args()
        if ret >= 0:
            return ret

        try:
            runner = self.instance.get_runner(args)
        except BaseException:  # pylint: disable=broad-except
            logger.exception("Exception raised while initializing the Cookbook %s:", self.full_name)
            return cookbook.CLASS_FAIL_INIT_RETCODE

        try:
            description = runner.runtime_description
        except BaseException:  # pylint: disable=broad-except
            logger.exception("Failed to get runtime_description from Cookbook %s:", self.full_name)
            description = ""

        _log.log_task_start(" ".join(("Cookbook", self.full_name, description)).strip())
        message = "raised while executing cookbook"

        try:
            raw_ret = runner.run()
            if raw_ret is None:
                ret = 0
            else:
                ret = raw_ret
        except KeyboardInterrupt:
            logger.error("Ctrl+c pressed")
            self._status = CookbookItem.error
            ret = cookbook.INTERRUPTED_RETCODE
        except SystemExit as e:
            if isinstance(e.code, int):
                ret = e.code
                if e.code == 0:
                    logger.info(
                        "SystemExit(0) %s %s, assuming success:",
                        message,
                        self.full_name,
                    )
                    self._status = CookbookItem.success
                else:
                    logger.exception("SystemExit(%d) %s %s:", e.code, message, self.full_name)
                    self._status = CookbookItem.error
            else:
                logger.exception("SystemExit('%s') %s %s:", e.code, message, self.full_name)
                self._status = CookbookItem.error
                ret = cookbook.EXCEPTION_RETCODE
        except BaseException:  # pylint: disable=broad-except
            logger.exception("Exception %s %s:", message, self.full_name)
            self._status = CookbookItem.failed
            ret = cookbook.EXCEPTION_RETCODE
        else:
            self._status = CookbookItem.success if ret == 0 else CookbookItem.failed

        if ret != 0:
            try:
                runner.rollback()
            except BaseException:  # pylint: disable=broad-except
                logger.exception("Exception %s %s rollback() (exit_code=%d):", message, self.full_name, ret)
                ret = cookbook.ROLLBACK_FAIL_RETCODE

        _log.log_task_end(
            self.status,
            f"Cookbook {self.full_name} (exit_code={ret}) {description}".strip(),
        )

        return ret

    def _parse_args(self) -> tuple[int, argparse.Namespace]:
        """Get the argument parser from the cookbook and parse the arguments.

        Returns:
            A 2-elements tuple with the return code to use and the parsed arguments. If the return code is different
            from ``-1`` it means that the cookbook should not be executed either because the help message was requested
            or the parse of the arguments failed or arguments were passed but the cookbook doesn't accept arguments.

        """
        args = argparse.Namespace()
        ret, parser = self._safe_call(
            self.instance.argument_parser,
            [],
            {},
            "raised while getting argument parser for cookbook",
            cookbook.GET_ARGS_PARSER_FAIL_RETCODE,
        )

        if ret >= 0 or parser is None:
            return ret, args

        return self._safe_call(
            parser.parse_args,
            [self.args],
            {},
            "raised while parsing arguments for cookbook",
            cookbook.PARSE_ARGS_FAIL_RETCODE,
        )

    def _safe_call(self, func: Callable, args: list, kwargs: dict, message: str, err_code: int) -> tuple[int, Any]:
        """Run any callable explicitly catching all exceptions including SystemExit, parsing the code of the latter.

        Arguments:
            func: the callable to call.
            args: positional arguments list to pass to the callable.
            kwargs: keyword arguments dictionary to pass to the callable.
            message: the message to log in case of exception.
            err_code: the error code to return in case of failure.

        Returns:
            A 2-element tuple with an integer for the return code as the first item and what the callable returned in
            the second item.

        """
        ret_code = -1
        ret_value = None
        try:
            ret_value = func(*args, **kwargs)
        except SystemExit as e:
            if isinstance(e.code, int):
                ret_code = e.code
            else:
                logger.exception("SystemExit('%s') %s %s:", e.code, message, self.full_name)
                ret_code = err_code
        except BaseException:  # pylint: disable=broad-except
            logger.exception("Exception %s %s:", message, self.full_name)
            ret_code = err_code

        return ret_code, ret_value


class TreeItem(BaseItem):
    """Tree of cookbook items class."""

    back_answer: str = "b"
    """Interactive menu answer to go back to the parent menu."""
    help_answer: str = "h"
    """Interactive menu answer to print the generic TreeItem help message."""
    quit_answer: str = "q"
    """Answer to quit the interactive menu."""
    help_message: str = HELP_MESSAGE.format(statuses=CookbookItem.statuses)
    """The generic TreeItem help message."""

    def __init__(
        self,
        module: _module_api.CookbooksModuleInterface,
        args: Sequence[str],
        spicerack: Spicerack,
        menu_title: str,
    ) -> None:
        """Override parent constructor to add menu-specific initialization.

        Arguments:
            module: a cookbook Python module.
            args: any sequence with the command line arguments to pass to the item.
            spicerack: the initialized instance of the library.
            menu_title: the title to use for the menu.

        """
        super().__init__(args, spicerack)
        self.module = module
        self.parent: Optional[TreeItem] = None
        self.menu_title = menu_title

        if "." in self.module.__name__:
            self.full_name = self.module.__name__.split(".", 1)[1]
        else:
            self.full_name = self.module.__name__

        if "." in self.full_name:
            self.path, self.name = self.full_name.rsplit(".", 1)
        else:
            self.name = self.full_name

    @property
    def status(self) -> str:
        """Getter for the menu status, returns a string representation of the status of its tasks."""
        completed, total = self.calculate_status()
        if completed == total:
            message = "DONE"
        else:
            message = f"{completed}/{total}"

        return message

    def append(self, item: BaseItem, add_parent: bool = True) -> None:
        """Append an item to this menu.

        Arguments:
            item: the item to append.
            add_parent: wheter to set the parent of the new item to the current instance.

        """
        if add_parent and isinstance(item, TreeItem):
            item.parent = self

        self.items[item.name] = item

    def run(self) -> int:
        """Execute the menu in an interactive way (infinite loop).

        Returns:
            Being an interactive menu it always returns 0.

        """
        try:
            while True:
                self.run_once()
        except StopIteration:
            pass

        return 0

    def show(self) -> None:
        """Print the menu to stdout."""
        for name in sorted(self.items.keys()):
            item = self.items[name]
            print(f"[{item.status}] {name}: {item.title}")

        if self.parent is None:
            print(f"{TreeItem.quit_answer} - Quit")
        else:
            print(f"{TreeItem.back_answer} - Back to parent menu")

        print(f"{TreeItem.help_answer} - Help")

    def calculate_status(self) -> tuple[int, int]:
        """Calculate the status of a menu, checking the status of all it's tasks recursively.

        Returns:
            A 2-elements tuple with the number of completed and total items.

        """
        completed: int = 0
        total: int = 0
        for item in self.items.values():
            if isinstance(item, TreeItem):
                sub_completed, sub_total = item.calculate_status()
                completed += sub_completed
                total += sub_total
            elif isinstance(item, CookbookItem):
                total += 1
                if item.status != CookbookItem.not_run:
                    completed += 1
            else:  # pragma: no cover | This should never happen
                raise MenuError(f"Unknown item of type {type(item)}")

        return completed, total

    def get_tree(self) -> str:
        """Return the tree representation of the menu as string."""
        lines = self.get_menu_tree(0, [])
        if not lines:
            return ""

        lines_str = "\n".join(lines)
        return f"{self.menu_title}\n{lines_str}\n"

    def get_menu_tree(self, level: int, cont_levels: list[bool]) -> list[str]:
        """Return the tree lines for a given menu.

        Arguments:
            level: how many levels the item is nested in the tree.
            cont_levels: a list of size levels with booleans that indicate for each level if the continuation prefix
                (:py:data:`True`) or an empty prefix (:py:data:`False`) should be used.

        """
        lines: list[str] = []
        for i, key in enumerate(sorted(self.items.keys(), key=lambda x: self.items[x].full_name)):
            is_final = i == len(self.items) - 1
            item = self.items[key]
            if isinstance(item, TreeItem) and not item.items:
                continue

            prefix = self._get_line_prefix(level, cont_levels, is_final)
            if self.spicerack.verbose:
                line = f"{prefix}{item.full_name}: {item.title}"
            else:
                line = f"{prefix}{item.full_name}"

            lines.append(line)

            if item.items:
                lines += cast(TreeItem, item).get_menu_tree(level + 1, cont_levels + [not is_final])

        return lines

    def run_once(self) -> None:
        """Run the menu in an interactive way."""
        print(f"#--- {self.verbose_title} args={self.args} ---#")
        self.show()

        if not sys.stdout.isatty():
            print("Not a tty, exiting.")
            raise StopIteration

        try:
            answer = input(">>> ")
        except (EOFError, KeyboardInterrupt) as e:
            print("QUIT")
            raise StopIteration from e  # Ctrl+d or Ctrl+c pressed while waiting for input

        if not answer:
            return

        if answer == TreeItem.help_answer:
            print(self.help_message)
            return

        if answer == TreeItem.quit_answer and self.parent is None:
            raise StopIteration

        if answer == TreeItem.back_answer and self.parent is not None:
            raise StopIteration

        name, *args = shlex.split(answer)
        if name not in self.items:
            print("==> Invalid input <==")
            return

        item = self.items[name]
        item.args = self._get_item_args(item, args)
        item.run()

    def _get_item_args(self, item: BaseItem, args: Sequence[str]) -> Sequence[str]:
        """Get the arguments to pass to the given item.

        Arguments:
            item: the item to pass the arguments to.
            args: any sequence with the arguments passed via interactive menu to this item.

        """
        if args:  # Override the item arguments with the ones passed interactively, if any.
            return args

        if self.args:  # Override the item arguments with the ones of the menu, if any.
            return self.args

        return item.args

    @property
    def title(self) -> str:
        """Returns the title of the instance item."""
        try:
            title = self.module.__title__.splitlines()[0]
        except AttributeError as e:
            logger.debug("Unable to detect title for module %s: %s", self.path, e)
            title = self.fallback_title

        return title



"""DNS Discovery module."""
import logging
from collections import defaultdict
from collections.abc import Iterator
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Optional, Union

import dns

# you might be tempted to remove this import. Don't, or dnspython won't work.
from dns import resolver
from dns.exception import DNSException

from spicerack.confctl import ConftoolEntity
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError

logger = logging.getLogger(__name__)


class DiscoveryError(SpicerackError):
    """Custom exception class for errors of the Discovery class."""


class DiscoveryCheckError(SpicerackCheckError):
    """Custom exception class for errors while performing checks."""


class Discovery:
    """Class to manage Confctl discovery objects."""

    def __init__(
        self,
        *,
        conftool: ConftoolEntity,
        authdns_servers: dict[str, str],
        records: list[str],
        dry_run: bool = True,
    ) -> None:
        """Initialize the instance.

        Arguments:
            conftool: the conftool instance for the discovery type objects.
            authdns_servers: a dictionary where keys are the hostnames and values are the IPs of the authoritative
                nameservers to be used.
            records: list of strings, each one must be a Discovery DNS record name.
            dry_run: whether this is a DRY-RUN.

        Raises:
            spicerack.dnsdisc.DiscoveryError: if unable to initialize the resolvers.

        """
        self._conftool = conftool
        self._records = records
        self._dry_run = dry_run

        self._resolvers: dict[str, resolver.Resolver] = {}
        for nameserver, nameserver_ip in authdns_servers.items():
            self._resolvers[nameserver] = resolver.Resolver(configure=False)
            self._resolvers[nameserver].port = 5353
            self._resolvers[nameserver].nameservers = [nameserver_ip]

    @property
    def _conftool_selector(self) -> str:
        """Returns the Conftool selector for the records."""
        regexp = "|".join(self._records)
        return f"({regexp})"

    @property
    def active_datacenters(self) -> defaultdict:
        """Information about pooled state of services.

        Returns:
            A map of services, with values given by a list of datacenters where the service is pooled, i.e.::

                {
                    'svc_foo': ['dc1', 'dc2'],
                    'svc_bar': ['dc1'],
                }

        """
        services = defaultdict(list)
        for obj in self._conftool.get(dnsdisc=self._conftool_selector):
            if obj.pooled:
                service = obj.tags["dnsdisc"]
                services[service].append(obj.name)

        return services

    def resolve_address(self, name: str) -> str:
        """Resolve the IP of a given record.

        Todo:
            move a more generalized version of this into a DNS resolver module.

        Arguments:
            name: the DNS record to resolve.

        Raises:
            spicerack.dnsdisc.DiscoveryError: if unable to resolve the address.

        """
        try:  # Querying the first resolver
            return next(iter(self._resolvers.values())).query(name)[0].address
        except DNSException as e:
            raise DiscoveryError(f"Unable to resolve {name}") from e

    def update_ttl(self, ttl: int) -> None:
        """Update the TTL for all registered records.

        Arguments:
            ttl: the new TTL value to set.

        Raises:
            spicerack.dnsdisc.DiscoveryError: if the check of the modified TTL fail and not in DRY-RUN mode.

        """
        # DRY-RUN handled by confctl
        logger.debug("Updating the TTL of %s to %d seconds", self._conftool_selector, ttl)
        self._conftool.set_and_verify("ttl", ttl, dnsdisc=self._conftool_selector)
        try:
            self.check_ttl(ttl)
        except DiscoveryCheckError:
            if not self._dry_run:
                raise

    @retry(
        tries=10,
        backoff_mode="constant",
        exceptions=(DiscoveryCheckError,),
        failure_message="Waiting for DNS TTL update...",
    )
    def check_ttl(self, ttl: int) -> None:
        """Check the TTL for all records.

        Arguments:
            ttl: the expected TTL value.

        Raises:
            spicerack.dnsdisc.DiscoveryError: if the expected TTL is not found.

        """
        logger.debug("Checking that TTL=%d for %s discovery.wmnet records", ttl, self._records)

        for record in self.resolve():
            if record.ttl != ttl:
                raise DiscoveryCheckError(f"Expected TTL '{ttl}', got '{record.ttl}' for record {record[0].address}")
        if len(self._records) == 1:
            logger.info("%s.discovery.wmnet TTL is correct.", self._records[0])
        else:
            logger.info("%s discovery.wmnet TTLs are correct.", self._records)

    @retry(
        tries=10,
        backoff_mode="constant",
        exceptions=(DiscoveryError,),
        failure_message="Waiting for DNS record update...",
    )
    def check_record(self, name: str, expected_name: str) -> None:
        """Check that a Discovery record resolves on all authoritative resolvers to the correct IP.

        The IP to use for the comparison is obtained resolving the expected_name record.
        For example with name='servicename-rw.discovery.wmnet' and expected_name='servicename.svc.eqiad.wmnet', this
        method will resolve the 'expected_name' to get its IP address and then verify that on all authoritative
        resolvers the record for 'name' resolves to the same IP.
        It is retried to allow the change to be propagated through all authoritative resolvers.

        See Also:
            https://wikitech.wikimedia.org/wiki/DNS/Discovery

        Arguments:
            name: the record to check the resolution for.
            expected_name: the name of a record to be resolved and used as the expected address.

        Raises:
            spicerack.dnsdisc.DiscoveryError: if the record doesn't match the IP of the expected_name.

        """
        expected_address = self.resolve_address(expected_name)
        logger.info(
            "Checking that %s.discovery.wmnet records matches %s (%s)",
            name,
            expected_name,
            expected_address,
        )

        failed = False
        for record in self.resolve(name=name):
            if not self._dry_run and record[0].address != expected_address:
                failed = True
                logger.error(
                    "Expected IP '%s', got '%s' for record %s",
                    expected_address,
                    record[0].address,
                    name,
                )

        if failed:
            raise DiscoveryError(f"Resolved record {name} with the wrong IP")

        logger.info("%s.discovery.wmnet record is correct.", name)

    def resolve(self, name: Optional[str] = None) -> Iterator[resolver.Answer]:
        """Generator that yields the resolved records.

        Todo:
            move a more generalized version of this into a DNS resolver module.

        Arguments:
            name: record name to use for the resolution instead of self.records.

        Yields:
            dns.resolver.Answer: the DNS response.

        Raises:
            spicerack.dnsdic.DiscoveryError: if unable to resolve the address.

        """
        if name is not None:
            records = [name]
        else:
            records = self._records

        for nameserver, dns_resolver in self._resolvers.items():
            for record in records:
                try:
                    record_name = f"{record}.discovery.wmnet"
                    answer = dns_resolver.query(record_name)
                except DNSException as e:
                    raise DiscoveryError(f"Unable to resolve {record_name} from {nameserver}") from e

                logger.debug(
                    "[%s] %s -> %s TTL %d",
                    nameserver,
                    record,
                    answer[0].address,
                    answer.ttl,
                )
                yield answer

    def pool(self, datacenter: str) -> None:
        """Set the records as pooled in the given datacenter.

        Arguments:
            datacenter: the DC in which to pool the discovery records.

        """
        # DRY-RUN handled by confctl
        self._conftool.set_and_verify("pooled", True, dnsdisc=self._conftool_selector, name=datacenter)

    def depool(self, datacenter: str) -> None:
        """Set the records as depooled in the given datacenter.

        Arguments:
            datacenter: the DC from which to depool the discovery records.

        """
        self.check_if_depoolable(datacenter)
        # DRY-RUN handled by confctl
        self._conftool.set_and_verify("pooled", False, dnsdisc=self._conftool_selector, name=datacenter)

    def check_if_depoolable(self, datacenter: str) -> None:
        """Determine if a datacenter can be depooled for all records.

        Arguments:
            datacenter: the datacenter to depool.

        Raises:
            spicerack.dnsdisc.DiscoveryError: if any service cannot be depooled.

        """
        # NB: we only discard services that would become inactive removing the current DC
        # we don't care about services that are completely down.
        non_depoolable = [svc for svc, dcs in self.active_datacenters.items() if dcs == [datacenter]]
        if non_depoolable:
            services = ", ".join(sorted(non_depoolable))
            message = f"Services {services} cannot be depooled as they are only active in {datacenter}"
            if self._dry_run:
                logger.debug(message)
            else:
                raise DiscoveryError(message)

    def resolve_with_client_ip(
        self, record: str, client_ip: Union[IPv4Address, IPv6Address]
    ) -> dict[str, Union[IPv4Address, IPv6Address]]:
        """Resolves a discovery record with a specific client IP and returns the resolved address grouped by nameserver.

        Arguments:
            record: record name to use for the resolution.
            client_ip: IP address to be used in EDNS client subnet.

        Raises:
            spicerack.discovery.DiscoveryError: if unable to resolve the address.

        """
        ips_by_ns: dict[str, Union[IPv4Address, IPv6Address]] = {}
        if record not in self._records:
            raise DiscoveryError(f"Record '{record}' not found")
        # Craft a query message
        record_name = dns.name.from_text(f"{record}.discovery.wmnet")
        rdtype = dns.rdatatype.from_text("A")
        ecs_option_client_ip = dns.edns.ECSOption(str(client_ip))
        query_msg = dns.message.make_query(record_name, rdtype)
        query_msg.use_edns(options=[ecs_option_client_ip])

        for nameserver, dns_resolver in self._resolvers.items():
            # Make the query. We catch generic exceptions as
            # dns.query.udp can raise many exceptions.
            try:
                query_response, _ = dns.query.udp_with_fallback(
                    query_msg, dns_resolver.nameservers[0], port=dns_resolver.port
                )
            except Exception as exc:
                raise DiscoveryError(f"Unable to resolve {record_name} from {nameserver}") from exc
            # Build an Answer instance as a Stub Resolver would
            try:
                response = resolver.Answer(record_name, rdtype, dns.rdataclass.from_text("IN"), query_response)
                # If no IN record was present in the response, the constructor raises an exception
                # so we don't need to check for the presence of at least one RRset in the Answer object.
                ips_by_ns[nameserver] = ip_address(response[0].address)
            except (DNSException, IndexError, StopIteration) as exc:
                raise DiscoveryError(f"Unable to resolve {record_name} from {nameserver}: {exc}") from exc
        return ips_by_ns



"""Cool feature external module."""


class CoolFeature:
    """An external module class."""

    def __init__(self, feature: str, *, dry_run: bool = True):
        """Instantiate the class."""
        self._feature = feature
        self._dry_run = dry_run

    def __str__(self) -> None:
        """String representation of the instance."""
        return f"{self._feature} is a cool feature!"



"""Exceptions module."""


class SpicerackError(Exception):
    """Parent exception class for all Spicerack exceptions."""


class SpicerackCheckError(SpicerackError):
    """Parent exception class for all Spicerack exceptions regarding checks.

    Particularly useful when some write action is performed and then checked but in dry-run mode, as in this mode the
    write action will not actually change anything and the check will then fail, but should be catchable separately
    from the other potential exceptions that could be raised.
    """



"""Constants module."""

PUPPET_CA_PATH: str = "/etc/ssl/certs/Puppet_Internal_CA.pem"
"""The path to the Puppet Signing CA cert"""

KEYHOLDER_SOCK: str = "/run/keyholder/proxy.sock"
"""The path to the keyholder agent sock file used for cumin remote commands"""



"""MysqlLegacy module tests."""
import logging
from datetime import datetime
from unittest import mock

import pytest
from cumin import Config, nodeset

from spicerack import mysql_legacy
from spicerack.remote import Remote, RemoteHosts
from spicerack.tests import get_fixture_path
from spicerack.tests.unit.test_remote import mock_cumin

EQIAD_CORE_MASTERS_QUERY = "db10[01-11]"
CODFW_CORE_MASTERS_QUERY = "db20[01-11]"


class TestMysqlLegacyRemoteHosts:
    """Test class for the MysqlLegacyRemoteHosts class."""

    @mock.patch("spicerack.remote.transports", autospec=True)
    def setup_method(self, _, mocked_transports):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        self.mocked_transports = mocked_transports
        self.mysql_remote_hosts = mysql_legacy.MysqlLegacyRemoteHosts(
            RemoteHosts(self.config, nodeset("host[1-9]"), dry_run=False)
        )
        self.expected = [(nodeset("host1"), "output1")]

    def test_execute(self):
        """Calling run_query() should run the given query in the target hosts."""
        mock_cumin(self.mocked_transports, 0)
        results = self.mysql_remote_hosts.run_query("query1")
        assert [(host, msg.message().decode()) for host, msg in results] == self.expected


class TestMysqlLegacy:
    """MysqlLegacy class tests."""

    @mock.patch("spicerack.remote.transports", autospec=True)
    def setup_method(self, _, mocked_transports):
        """Initialize the test environment for MysqlLegacy."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        self.mocked_transports = mocked_transports
        self.mocked_remote = mock.MagicMock(spec_set=Remote)
        self.mysql = mysql_legacy.MysqlLegacy(self.mocked_remote, dry_run=False)

    def test_get_dbs(self):
        """It should return and instance of MysqlLegacyRemoteHosts for the matching hosts."""
        self.mysql.get_dbs("query")
        self.mocked_remote.query.assert_called_once_with("query")

    @pytest.mark.parametrize(
        "kwargs, query, match",
        (
            ({}, "A:db-core", "db10[01-99],db20[01-99]"),
            ({"datacenter": "eqiad"}, "A:db-core and A:eqiad", "db10[01-99]"),
            (
                {"section": "s1"},
                "A:db-core and A:db-section-s1",
                "db10[01-10],db20[01-10]",
            ),
            (
                {"replication_role": "master"},
                "A:db-core and A:db-role-master",
                ",".join([EQIAD_CORE_MASTERS_QUERY, CODFW_CORE_MASTERS_QUERY]),
            ),
            (
                {"datacenter": "eqiad", "section": "s1"},
                "A:db-core and A:eqiad and A:db-section-s1",
                "db10[01-10]",
            ),
            (
                {"datacenter": "eqiad", "replication_role": "master"},
                "A:db-core and A:eqiad and A:db-role-master",
                EQIAD_CORE_MASTERS_QUERY,
            ),
            (
                {"datacenter": "eqiad", "replication_role": "master", "excludes": ("s2",)},
                "A:db-core and A:eqiad and not A:db-section-s2 and A:db-role-master",
                "db10[01-10]",
            ),
            (
                {"section": "s1", "replication_role": "master"},
                "A:db-core and A:db-section-s1 and A:db-role-master",
                "db1001,db2001",
            ),
            (
                {"datacenter": "eqiad", "section": "s1", "replication_role": "master"},
                "A:db-core and A:eqiad and A:db-section-s1 and A:db-role-master",
                "db1001",
            ),
        ),
    )
    def test_get_core_dbs_ok(self, kwargs, query, match):
        """It should return the right DBs based on the parameters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset(match))
        self.mysql.get_core_dbs(**kwargs)
        self.mocked_remote.query.assert_called_once_with(query)

    @pytest.mark.parametrize(
        "kwargs",
        (
            {"datacenter": "invalid"},
            {"section": "invalid"},
            {"replication_role": "invalid"},
            {"excludes": ("invalid",)},
        ),
    )
    def test_get_core_dbs_fail(self, kwargs):
        """It should raise MysqlLegacyError if called with invalid parameters."""
        key = list(kwargs.keys())[0]
        message = f"Got invalid {key}"
        with pytest.raises(mysql_legacy.MysqlLegacyError, match=message):
            self.mysql.get_core_dbs(**kwargs)

        assert not self.mocked_remote.query.called

    def test_get_core_dbs_fail_sanity_check(self):
        """It should raise MysqlLegacyError if matching an invalid number of hosts when looking for masters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Matched 1 masters, expected 11"):
            self.mysql.get_core_dbs(datacenter="eqiad", replication_role="master")

        assert self.mocked_remote.query.called

    @pytest.mark.parametrize("mode, value", (("readonly", b"1"), ("readwrite", b"0")))
    def test_set_core_masters_readonly(self, mode, value, caplog):
        """It should set the masters as read-only/read-write."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(self.mocked_transports, 0, retvals=[[("db10[01-11]", value)]])
        with caplog.at_level(logging.DEBUG):
            getattr(self.mysql, "set_core_masters_" + mode)("eqiad")
        assert "SET GLOBAL read_only=" + value.decode() in caplog.text

    @pytest.mark.parametrize("readonly, reply", ((True, b"1"), (False, b"0")))
    def test_verify_core_masters_readonly_ok(self, readonly, reply, caplog):
        """Should verify that the masters have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(self.mocked_transports, 0, retvals=[[("db10[01-11]", reply)]])
        with caplog.at_level(logging.DEBUG):
            self.mysql.verify_core_masters_readonly("eqiad", readonly)
        assert "SELECT @@global.read_only" in caplog.text

    def test_verify_core_masters_readonly_fail(self):
        """Should raise MysqlLegacyError if some masters do not have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[("db1001", b"0"), ("db10[02-11]", b"1")]],
        )
        with pytest.raises(
            mysql_legacy.MysqlLegacyError,
            match="Verification failed that core DB masters",
        ):
            self.mysql.verify_core_masters_readonly("eqiad", True)

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_ok(self, mocked_sleep):
        """Should check that all core masters are in sync with the master in the other DC."""
        hosts = nodeset(EQIAD_CORE_MASTERS_QUERY)
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, nodeset(host)) for host in hosts] * 2
        retvals = [[(host, b"2018-09-06T10:00:00.000000")] for host in hosts]  # first heartbeat
        retvals += [[(host, b"2018-09-06T10:00:01.000000")] for host in hosts]  # second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        self.mysql.check_core_masters_in_sync("eqiad", "codfw")
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_fail_heartbeat(self, mocked_sleep):
        """Should raise MysqlLegacyError if unable to get the heartbeat from the current master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to get heartbeat from master"):
            self.mysql.check_core_masters_in_sync("eqiad", "codfw")
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_not_in_sync(self, mocked_sleep):
        """Should raise MysqlLegacyError if a master is not in sync with the one in the other DC."""
        hosts = nodeset(EQIAD_CORE_MASTERS_QUERY)
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, nodeset(host)) for host in hosts] + [
            RemoteHosts(self.config, nodeset("db1001"))
        ] * 3
        retvals = [[(host, b"2018-09-06T10:00:00.000000")] for host in hosts]  # first heartbeat
        retvals += [[("db1001", b"2018-09-06T10:00:00.000000")]] * 3  # 3 failed retries of second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        with pytest.raises(
            mysql_legacy.MysqlLegacyError,
            match=r"Heartbeat from master db1001 for section .* not yet in sync",
        ):
            self.mysql.check_core_masters_in_sync("eqiad", "codfw")

        assert mocked_sleep.called

    def test_get_core_masters_heartbeats_wrong_data(self):
        """Should raise MysqlLegacyError if unable to convert the heartbeat into a datetime."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[("db1001", b"2018-09-06-10:00:00.000000")]],
        )
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to convert heartbeat"):
            self.mysql.get_core_masters_heartbeats("eqiad", "codfw")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_heartbeats_fail(self, mocked_sleep):
        """Should raise MysqlLegacyError if unable to get the heartbeat from the master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to get heartbeat from master"):
            self.mysql.check_core_masters_heartbeats("eqiad", "codfw", {"s1": datetime.utcnow()})

        assert mocked_sleep.called



"""MediaWiki module tests."""
import json
from unittest import mock

import pytest
import requests

from spicerack.mediawiki import MediaWiki, MediaWikiCheckError, MediaWikiError
from spicerack.remote import RemoteExecutionError


class TestMediaWiki:
    """MediaWiki class tests."""

    def setup_method(self):
        """Initialize the test environment for MediaWiki."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_confctl = mock.MagicMock()
        self.mocked_remote = mock.MagicMock()
        self.mocked_remote.query.return_value.hosts = ["host1"]
        self.username = "user1"
        self.siteinfo_url = "https://api.svc.eqiad.wmnet/w/api.php"
        self.siteinfo_rw = {
            "batchcomplete": True,
            "query": {
                "general": {
                    "readonly": False,
                    "wmf-config": {
                        "wmfEtcdLastModifiedIndex": 123456,
                        "wmfMasterDatacenter": "eqiad",
                    },
                },
            },
        }
        self.siteinfo_ro = {
            "batchcomplete": True,
            "query": {
                "general": {
                    "readonly": True,
                    "readonlyreason": "MediaWiki is in read-only mode for maintenance.",
                    "wmf-config": {
                        "wmfEtcdLastModifiedIndex": 123456,
                        "wmfMasterDatacenter": "eqiad",
                    },
                },
            },
        }

        self.mediawiki = MediaWiki(self.mocked_confctl, self.mocked_remote, self.username, dry_run=False)
        self.mediawiki_dry_run = MediaWiki(self.mocked_confctl, self.mocked_remote, self.username)

    def test_check_config_line(self, requests_mock):
        """It should verify the config published at noc.wikimedia.org."""
        requests_mock.get("http://host1/conf/file1.php.txt", text="data")
        assert self.mediawiki.check_config_line("file1", "data") is True

    def test_get_master_datacenter(self):
        """It should return athe primary site."""
        self.mocked_confctl.get.return_value.__next__.return_value.val = "dc1"
        dc = self.mediawiki.get_master_datacenter()
        self.mocked_confctl.get.assert_called_once_with(scope="common", name="WMFMasterDatacenter")
        assert dc == "dc1"

    def test_get_siteinfo(self, requests_mock):
        """It should get the siteinfo API from a canary host."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        assert self.mediawiki.get_siteinfo("eqiad") == self.siteinfo_ro

    def test_check_siteinfo_ok(self, requests_mock):
        """It should check that a specific key in siteinfo API matches a value and not raise exception."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_rw))
        self.mediawiki.check_siteinfo(
            "eqiad",
            {("query", "general", "wmf-config", "wmfEtcdLastModifiedIndex"): 123456},
        )

    def test_check_siteinfo_multi_ok(self, requests_mock):
        """It should check that multiple keys in siteinfo API matches multiple values and not raise exception."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_rw))
        self.mediawiki.check_siteinfo(
            "eqiad",
            {
                ("query", "general", "wmf-config", "wmfEtcdLastModifiedIndex"): 123456,
                ("batchcomplete",): True,
                ("query", "general", "readonly"): False,
            },
        )

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_siteinfo_raise_check(self, mocked_sleep, requests_mock):
        """It should retry if it doesn't match and raise MediaWikiCheckError after all retries have failed."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(
            MediaWikiCheckError,
            match=r"Expected 'invalid', got 'True' for path: \('batchcomplete',\)",
        ):
            self.mediawiki.check_siteinfo("eqiad", {("batchcomplete",): "invalid"})

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_siteinfo_raise_error(self, mocked_sleep, requests_mock):
        """It should retry if it fails and raise MediaWikiError after all retries have failed."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiError, match="Failed to traverse siteinfo for key invalid"):
            self.mediawiki.check_siteinfo("eqiad", {("invalid",): "invalid"})

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_siteinfo_key_error(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiError if the key is not present in siteinfo."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiError, match="Failed to traverse siteinfo for key invalid"):
            self.mediawiki.check_siteinfo("eqiad", {("invalid",): "invalid"})

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_siteinfo_timeout(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiError if it fails to get siteinfo."""
        requests_mock.get(self.siteinfo_url, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(MediaWikiError, match="Failed to get siteinfo"):
            self.mediawiki.check_siteinfo("eqiad", {("batchcomplete",): "invalid"})

        assert mocked_sleep.called

    def test_scap_sync_config_file(self):
        """It should run scap sync file on the deployment host."""
        self.mediawiki.scap_sync_config_file("file1", "deployed file1")
        self.mocked_remote.query.assert_called_once_with("C:Deployment::Rsync and R:Class%cron_ensure = absent")
        assert "scap sync-file" in self.mocked_remote.query.return_value.run_sync.call_args[0][0]

    def test_set_readonly(self, requests_mock):
        """It should set the readonly message in Conftool and verify it in siteinfo."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        message = self.siteinfo_ro["query"]["general"]["readonlyreason"]
        self.mediawiki.set_readonly("eqiad", message)
        self.mocked_confctl.set_and_verify.assert_called_once_with("val", message, name="ReadOnly", scope="eqiad")

    def test_set_readonly_dry_run(self, requests_mock):
        """It should not set the readonly message in Conftool and the siteinfo check should not raise."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        self.mediawiki_dry_run.set_readonly("eqiad", "invalid")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_set_readonly_raise(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiCheckError if unable to check the value has changed and not in dry_run."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiCheckError, match="Expected 'invalid', got"):
            self.mediawiki.set_readonly("eqiad", "invalid")

        assert mocked_sleep.called

    def test_set_readwrite(self, requests_mock):
        """It should set the readonly variale in Conftool to False and verify it in siteinfo."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_rw))
        self.mediawiki.set_readwrite("eqiad")
        self.mocked_confctl.set_and_verify.assert_called_once_with("val", False, name="ReadOnly", scope="eqiad")

    def test_set_readwrite_dry_run(self, requests_mock):
        """It should not set the readonly variable in Conftool and the siteinfo check should not raise."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        self.mediawiki_dry_run.set_readwrite("eqiad")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_set_readwrite_raise(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiCheckError if unable to check the value has changed and not in dry_run."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiCheckError, match="Expected 'False', got 'True'"):
            self.mediawiki.set_readwrite("eqiad")

        assert mocked_sleep.called

    def test_set_master_datacenter(self, requests_mock):
        """It should set the master datacenter in Conftool."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        requests_mock.get(
            self.siteinfo_url.replace("eqiad", "codfw"),
            text=json.dumps(self.siteinfo_ro),
        )
        self.mediawiki.set_master_datacenter("eqiad")
        self.mocked_confctl.set_and_verify.assert_called_once_with(
            "val", "eqiad", name="WMFMasterDatacenter", scope="common"
        )

    def test_set_master_datacenter_dry_run(self, requests_mock):
        """It should not set the master datacenter in Conftool and not raise as the siteinfo check fails."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        requests_mock.get(
            self.siteinfo_url.replace("eqiad", "codfw"),
            text=json.dumps(self.siteinfo_ro),
        )
        self.mediawiki_dry_run.set_master_datacenter("codfw")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_set_master_datacenter_raise(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiCheckError if unable to check the value has changed and not in dry_run."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        requests_mock.get(
            self.siteinfo_url.replace("eqiad", "codfw"),
            text=json.dumps(self.siteinfo_ro),
        )
        with pytest.raises(MediaWikiCheckError, match="Expected 'codfw', got 'eqiad'"):
            self.mediawiki.set_master_datacenter("codfw")

        assert mocked_sleep.called

    def test_check_periodic_jobs_enabled(self):
        """It should ensure that the periodic jobs are present and not commented out."""
        self.mediawiki.check_periodic_jobs_enabled("dc1")
        self.mocked_remote.query.assert_called_with("A:mw-maintenance and A:dc1")
        assert "systemctl list-units" in self.mocked_remote.query.return_value.run_sync.call_args[0][0]

    def test_stop_periodic_jobs(self):
        """It should ensure that the periodic jobs are stopped and remaining processes are killed."""
        self.mediawiki.stop_periodic_jobs("dc1")
        self.mocked_remote.query.assert_called_with("A:mw-maintenance and A:dc1")
        assert "systemctl stop" in self.mocked_remote.query.return_value.run_async.call_args_list[0][0][0]
        assert "killall" in self.mocked_remote.query.return_value.run_async.call_args_list[1][0][-2].command
        assert "systemctl list-units" in self.mocked_remote.query.return_value.run_async.call_args_list[2][0][0].command

    def test_stop_periodic_jobs_stray_procs(self):
        """It should not fail is there are leftover php stray processes."""
        self.mocked_remote.query.return_value.run_sync.side_effect = [
            RemoteExecutionError(10, "failed", iter(())),
        ]
        self.mediawiki.stop_periodic_jobs("dc1")



"""Puppet module."""
import json
import logging
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime, timedelta
from subprocess import CalledProcessError, check_output
from typing import Optional, Union, cast

from cumin import NodeSet, nodeset
from cumin.transports import Command

from spicerack.administrative import Reason
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError
from spicerack.remote import RemoteExecutionError, RemoteHosts, RemoteHostsAdapter

PUPPET_COMMON_SCRIPT: str = "/usr/local/share/bash/puppet-common.sh"
"""The absolute path of the puppet-common script shipped by Puppet with useful functions."""
logger = logging.getLogger(__name__)


def get_puppet_ca_hostname() -> str:
    """Return the FQDN of the current Puppet CA server.

    Raises:
        spicerack.puppet.PuppetMasterError: if unable to get the configured Puppet CA server.

    """
    try:
        output = (
            check_output(["puppet", "config", "print", "--section", "agent", "ca_server"]).decode().strip()  # nosec
        )
    except CalledProcessError as e:
        raise PuppetMasterError(f"Get Puppet ca_server failed (exit={e.returncode}): {e.output}") from e

    if not output:
        raise PuppetMasterError("Got empty ca_server from Puppet agent")

    return output


class PuppetHostsError(SpicerackError):
    """Custom base exception class for errors in the PuppetHosts class."""


class PuppetHostsCheckError(SpicerackCheckError):
    """Custom base exception class for check errors in the PuppetHosts class."""


class PuppetMasterError(SpicerackError):
    """Custom base exception class for errors in the PuppetMaster class."""


class PuppetMasterCheckError(SpicerackCheckError):
    """Custom exception class for check errors in the PuppetMaster class."""


class PuppetHosts(RemoteHostsAdapter):
    """Class to manage Puppet on the target hosts."""

    @contextmanager
    def disabled(self, reason: Reason, verbatim_reason: bool = False) -> Iterator[None]:
        """Context manager to perform actions while puppet is disabled.

        Arguments:
            reason: the reason to set for the Puppet disable and to use for the Puppet enable.
            verbatim_reason: if true use the reason value verbatim.

        """
        self.disable(reason, verbatim_reason)
        try:
            yield
        finally:
            self.enable(reason, verbatim_reason)

    def get_ca_servers(self) -> dict[str, str]:
        """Retrieve the ca_servers for each node."""
        raw_results = self._remote_hosts.run_sync("puppet config --section agent print ca_server")

        host_to_ca_server: dict[str, str] = {}
        for node_set, output in raw_results:
            for hostname in node_set:
                host_to_ca_server[hostname] = list(output.lines())[-1].decode()

        return host_to_ca_server

    @staticmethod
    def _puppet_reason(reason: Reason, verbatim_reason: bool = False) -> str:
        """Return a correctly quoted puppet message.

        Arguments:
            reason: the reason to set for the Puppet disable.
            verbatim_reason: if true use the reason value verbatim.

        """
        if verbatim_reason:
            return f'"{reason.reason}"'
        return reason.quoted()

    def disable(self, reason: Reason, verbatim_reason: bool = False) -> None:
        """Disable puppet with a specific reason.

        If Puppet was already disabled on a host with a different reason, the reason will not be overriden, allowing to
        leave Puppet disabled when re-enabling it if it was already disabled.

        Arguments:
            reason: the reason to set for the Puppet disable.
            verbatim_reason: if true use the reason value verbatim.

        """
        logger.info(
            "Disabling Puppet with reason %s on %d hosts: %s",
            reason.quoted(),
            len(self),
            self,
        )
        self._remote_hosts.run_sync("disable-puppet " + self._puppet_reason(reason, verbatim_reason))

    def enable(self, reason: Reason, verbatim_reason: bool = False) -> None:
        """Enable Puppet with a specific reason, it must be the same used to disable it.

        Puppet will be re-enabled only if it was disable with the same reason. If it was disable with a different reason
        it will keep being disabled.

        Arguments:
            reason: the reason to use for the Puppet enable.
            verbatim_reason: if true use the reason value verbatim.

        """
        logger.info(
            "Enabling Puppet with reason %s on %d hosts: %s",
            reason.quoted(),
            len(self),
            self,
        )
        self._remote_hosts.run_sync("enable-puppet " + self._puppet_reason(reason, verbatim_reason))

    def check_enabled(self) -> None:
        """Check if Puppet is enabled on all hosts.

        Raises:
            spicerack.puppet.PuppetHostsCheckError: if Puppet is disabled on some hosts.

        """
        disabled = self._get_disabled()[True]
        if disabled:
            raise PuppetHostsCheckError(f"Puppet is not enabled on those hosts: {disabled}")

    def check_disabled(self) -> None:
        """Check if Puppet is disabled on all hosts.

        Raises:
            spicerack.puppet.PuppetHostsCheckError: if Puppet is enabled on some hosts.

        """
        enabled = self._get_disabled()[False]
        if enabled:
            raise PuppetHostsCheckError(f"Puppet is not disabled on those hosts: {enabled}")

    def run(  # pylint: disable=too-many-arguments
        self,
        timeout: int = 300,
        enable_reason: Optional[Reason] = None,
        quiet: bool = False,
        failed_only: bool = False,
        force: bool = False,
        attempts: int = 0,
        batch_size: int = 10,
    ) -> None:
        """Run Puppet.

        Arguments:
            timeout: the timeout in seconds to set in Cumin for the execution of the command.
            enable_reason: the reason to use to contextually re-enable Puppet if it was disabled.
            quiet: suppress Puppet output if True.
            failed_only: run Puppet only if the last run failed.
            force: forcely re-enable Puppet if it was disabled with ANY message.
            attempts: override the default number of attempts waiting that an in-flight Puppet run completes before
                timing out as set in run-puppet-agent.
            batch_size: how many concurrent Puppet runs to perform. The default value is tailored to not overload the
                Puppet masters.

        """
        args = []
        if enable_reason is not None:
            args += ["--enable", enable_reason.quoted()]
        if quiet:
            args.append("--quiet")
        if failed_only:
            args.append("--failed-only")
        if force:
            args.append("--force")
        if attempts:
            args += ["--attempts", str(attempts)]

        args_string = " ".join(args)
        command = f"run-puppet-agent {args_string}"
        logger.info("Running Puppet with args %s on %d hosts: %s", args_string, len(self), self)
        self._remote_hosts.run_sync(Command(command, timeout=timeout), batch_size=batch_size)

    def first_run(self, has_systemd: bool = True) -> Iterator[tuple]:
        """Perform the first Puppet run on a clean host without using custom wrappers.

        Arguments:
            has_systemd: if the host has systemd as init system.

        """
        commands = []
        if has_systemd:
            commands += [
                "systemctl stop puppet.service",
                "systemctl reset-failed puppet.service || true",
            ]

        commands += [
            "puppet agent --enable",
            Command(
                (
                    "puppet agent --onetime --no-daemonize --verbose --no-splay --show_diff --ignorecache "
                    "--no-usecacheonfailure"
                ),
                timeout=10800,
            ),
        ]

        logger.info("Starting first Puppet run (sit back, relax, and enjoy the wait)")
        results = self._remote_hosts.run_sync(*commands, print_output=False, print_progress_bars=False)
        logger.info("First Puppet run completed")
        return results

    def regenerate_certificate(self) -> dict[str, str]:
        """Delete the local Puppet certificate and generate a new CSR.

        Returns:
            A dictionary with hostnames as keys and CSR fingerprint as values.

        """
        logger.info("Deleting local Puppet certificate on %d hosts: %s", len(self), self)
        self._remote_hosts.run_sync("rm -rfv /var/lib/puppet/ssl")

        fingerprints = {}
        errors = []
        # The return codes for the cert generation are not well defined, we'll
        # check if it worked by searching for the fingerprint and parsing the
        # output.
        command = Command("puppet agent --test --color=false", ok_codes=[])
        logger.info("Generating a new Puppet certificate on %d hosts: %s", len(self), self)
        for node_set, output in self._remote_hosts.run_sync(command, print_output=False):
            for line in output.message().decode().splitlines():
                if line.startswith("Error:"):
                    errors.append((node_set, line))
                    continue

                if "Certificate Request fingerprint" not in line:
                    continue

                fingerprint = ":".join(line.split(":")[2:]).strip()
                if not fingerprint:
                    continue

                logger.info("Generated CSR for host %s: %s", node_set, fingerprint)
                for host in node_set:
                    fingerprints[host] = fingerprint

        if len(fingerprints) != len(self):
            raise PuppetHostsError(
                "Unable to find CSR fingerprints for all hosts, detected errors are:\n"
                + "\n".join(f"{node_set}: {line}" for node_set, line in errors)
            )

        return fingerprints

    def wait(self) -> None:
        """Wait until the next successful Puppet run is completed."""
        self.wait_since(datetime.utcnow())

    @retry(
        tries=60,
        delay=timedelta(seconds=30),
        backoff_mode="linear",
        exceptions=(PuppetHostsCheckError,),
    )
    def wait_since(self, start: datetime) -> None:
        """Wait until a successful Puppet run is completed after the start time.

        Arguments:
            start: wait until a Puppet run is completed after this time.

        Raises:
            spicerack.puppet.PuppetHostsCheckError: if unable to get a successful Puppet run within the timeout.

        """
        remaining_nodes = self._remote_hosts.hosts
        command = (
            f"source {PUPPET_COMMON_SCRIPT} && last_run_success && "
            "awk /last_run/'{ print $2 }' \"${PUPPET_SUMMARY}\""
        )

        logger.debug("Polling the completion of a successful Puppet run")
        try:
            for node_set, output in self._remote_hosts.run_sync(
                command, is_safe=True, print_output=False, print_progress_bars=False
            ):
                last_run = datetime.utcfromtimestamp(int(output.message().decode()))
                if last_run <= start:
                    raise PuppetHostsCheckError(f"Successful Puppet run too old ({last_run} <= {start}) on: {node_set}")

                remaining_nodes.difference_update(node_set, strict=False)

        except RemoteExecutionError as e:
            raise PuppetHostsCheckError("Unable to find a successful Puppet run") from e

        if remaining_nodes:
            raise PuppetHostsCheckError(f"Unable to get successful Puppet run from: {remaining_nodes}")

        logger.info("Successful Puppet run found")

    def _get_disabled(self) -> dict[bool, NodeSet]:
        """Check if Puppet is disabled on the hosts.

        Returns:
            A dict with :py:class:`bool` keys for Puppet disabled or not and hosts
            :py:class:`ClusterShell.NodeSet.NodeSet` as values.

        """
        results = self._remote_hosts.run_sync(
            f'source {PUPPET_COMMON_SCRIPT} && test -f "${{PUPPET_DISABLEDLOCK}}" && echo "1" || echo "0"',
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

        disabled = {True: nodeset(), False: nodeset()}
        for node_set, output in results:
            result = bool(int(output.message().decode().strip()))
            disabled[result] |= node_set

        return disabled


class PuppetMaster(RemoteHostsAdapter):
    """Class to manage nodes and certificates on a Puppet master and Puppet CA server."""

    PUPPET_CERT_STATE_REQUESTED: str = "requested"
    """Puppet CA certificate status when requested."""
    PUPPET_CERT_STATE_SIGNED: str = "signed"
    """Puppet CA certificate status when signed."""

    def __init__(self, master_host: RemoteHosts) -> None:
        """Initialize the instance.

        Arguments:
            master_host: the remote hosts instance for the Puppetmaster and Puppet CA server. It must have only one
                target host.

        Raises:
            spicerack.puppet.PuppetMasterError: if the master_host doesn't have only one target host.

        """
        if len(master_host) != 1:
            raise PuppetMasterError(
                f"The master_host instance must target only one host, got {len(master_host)}: {master_host}"
            )
        super().__init__(master_host)

    @property
    def master_host(self) -> RemoteHosts:
        """Accessor for the master_host property."""
        return self._remote_hosts

    def delete(self, hostname: str) -> None:
        """Remove the host from the Puppet master and PuppetDB.

        Clean up signed certs, cached facts, node objects, and reports in the Puppet master, deactivate it in PuppetDB.
        Doesn't raise exception if the host was already removed.

        Arguments:
            hostname: the FQDN of the host for which to remove the certificate.

        """
        commands = [f"puppet node {action} {hostname}" for action in ("clean", "deactivate")]
        self.master_host.run_sync(*commands, print_progress_bars=False)

    def destroy(self, hostname: str) -> None:
        """Remove the certificate for the given hostname.

        If there is no certificate to remove it doesn't raise exception as the Puppet CA just outputs
        'Nothing was deleted'.

        Arguments:
            hostname: the FQDN of the host for which to remove the certificate.

        """
        self.master_host.run_sync(
            f"puppet ca --disable_warnings deprecations destroy {hostname}", print_progress_bars=False
        )

    def verify(self, hostname: str) -> None:
        """Verify that there is a valid certificate signed by the Puppet CA for the given hostname.

        Arguments:
            hostname: the FQDN of the host for which to verify the certificate.

        Raises:
            spicerack.puppet.PuppetMasterError: if the certificate is not valid.

        """
        response = cast(
            dict,
            self._run_json_command(f"puppet ca --disable_warnings deprecations --render-as json verify {hostname}"),
        )

        if not response["valid"]:
            raise PuppetMasterError(f"Invalid certificate for {hostname}: {response['error']}")

    def sign(self, hostname: str, fingerprint: str, allow_alt_names: bool = False) -> None:
        """Sign a CSR on the Puppet CA for the given host checking its fingerprint.

        Arguments:
            hostname: the FQDN of the host for which to sign the certificate.
            fingerprint: the fingerprint of the CSR generated on the client to verify it.
            allow_alt_names: whether to allow DNS alternative names in the certificate.

        Raises:
            spicerack.puppet.PuppetMasterError: if the certificate is in an unexpected state.

        """
        cert = self.get_certificate_metadata(hostname)
        if cert["state"] != PuppetMaster.PUPPET_CERT_STATE_REQUESTED:
            raise PuppetMasterError(f"Certificate for {hostname} not in requested state, got: {cert['state']}")

        if cert["fingerprint"] != fingerprint:
            raise PuppetMasterError(
                f"CSR fingerprint {cert['fingerprint']} for {hostname} does not match provided fingerprint "
                f"{fingerprint}"
            )

        if allow_alt_names:
            dns_option = "--allow-dns-alt-names"
        else:
            dns_option = "--no-allow-dns-alt-names"

        command = f"puppet cert --disable_warnings deprecations sign {dns_option} {hostname}"
        logger.info("Signing CSR for %s with fingerprint %s", hostname, fingerprint)
        executed = self.master_host.run_sync(command, print_output=False, print_progress_bars=False)

        cert = self.get_certificate_metadata(hostname)
        if cert["state"] != PuppetMaster.PUPPET_CERT_STATE_SIGNED:
            for _, output in executed:
                logger.error(output.message().decode())

            raise PuppetMasterError(f"Expected certificate for {hostname} to be signed, got: {cert['state']}")

    @retry(
        tries=10,
        delay=timedelta(seconds=5),
        backoff_mode="power",
        exceptions=(PuppetMasterCheckError,),
    )
    def wait_for_csr(self, hostname: str) -> None:
        """Poll until a CSR appears for the given hostname or the timeout is reached.

        Arguments:
            hostname: the FQDN of the host for which to check a CSR.

        Raises:
            spicerack.puppet.PuppetMasterError: if the certificate is in an unexpected state.
            spicerack.puppet.PuppetMasterCheckError: if within the timeout no CSR is found.

        """
        state = self.get_certificate_metadata(hostname)["state"]
        if state != PuppetMaster.PUPPET_CERT_STATE_REQUESTED:
            raise PuppetMasterError(f"Expected certificate in requested state, got: {state}")

    def get_certificate_metadata(self, hostname: str) -> dict:
        """Return the metadata of the certificate of the given hostname in the Puppet CA.

        Arguments:
            hostname: the FQDN of the host for which to verify the certificate.

        Returns:
            As returned by the Puppet CA CLI with the render as JSON option set. As example::

                {
                    'dns_alt_names': ['DNS:service.example.com'],
                    'fingerprint': '00:FF:...',
                    'fingerprints': {
                        'SHA1': '00:FF:...',
                        'SHA256': '00:FF:...',
                        'SHA512': '00:FF:...',
                        'default': '00:FF:...',
                    },
                    'name': 'host.example.com',
                    'state': 'signed',
                }

        Raises:
            spicerack.puppet.PuppetMasterCheckError: if no certificate is found.
            spicerack.puppet.PuppetMasterError: if more than one certificate is found or it has invalid data.

        """
        pattern = hostname.replace(".", r"\.")
        response = self._run_json_command(
            f'puppet ca --disable_warnings deprecations --render-as json list --all --subject "^{pattern}$"'
        )

        if not response:
            raise PuppetMasterCheckError(f"No certificate found for hostname: {hostname}")

        if len(response) > 1:
            raise PuppetMasterError(f"Expected one result from Puppet CA, got {len(response)}")

        metadata = response[0]
        if metadata["name"] != hostname:
            raise PuppetMasterError(f"Hostname mismatch {metadata['name']} != {hostname}")

        return metadata

    def _run_json_command(self, command: str) -> Union[dict, list]:
        """Execute and parse a Puppet CLI command that output JSON format.

        The commands run are assumed to be safe as the JSON format is useful for read-only operations only.

        Arguments:
            command: the command to execute on the Puppet master that returns JSON output.

        Raises:
            spicerack.puppet.PuppetMasterError: if unable to get or parse the command output.

        """
        for _, output in self.master_host.run_sync(
            command, is_safe=True, print_output=False, print_progress_bars=False
        ):
            lines = output.message().decode()
            break
        else:
            raise PuppetMasterError(f"Got no output from Puppet master while executing command: {command}")

        try:
            response = json.loads(lines)
        except ValueError as e:
            raise PuppetMasterError(f'Unable to parse Puppet master response for command "{command}": {lines}') from e

        return response



"""Administrative module."""
from typing import Optional

from spicerack.exceptions import SpicerackError


class ReasonError(SpicerackError):
    """Custom exception class for errors of the Reason class."""


class Reason:
    """Class to manage the reason for administrative actions."""

    def __init__(self, reason: str, username: str, hostname: str, *, task_id: Optional[str] = None) -> None:
        """Initialize the instance.

        Arguments:
            reason: the reason to use to justify an administrative action. The username and the hostname where the
                action was originated will be added to the reason automatically. The reason is meant to be passed to
                remote execution in double quotes, allowing to use Bash variables, if needed. Therefore the reason
                cannot contain double quotes.
            username: the username to mention in the reason as the author of the action.
            hostname: the hostname to mention in the reason as the host originating the action.
            task_id: the task ID to mention in the reason.

        Raises:
            spicerack.administrative.ReasonError: if any parameter contains double quotes.

        """
        self._reason = reason
        self._username = username
        self._hostname = hostname
        self._task_id = task_id

    def __setattr__(self, name: str, value: str) -> None:
        """Set an instance attribute with validation.

        Parameters as required by Python's data model, see `object.__setattr__`.

        Raises:
            ReasonError: on validation error of the parameters.

        """
        if value is not None and '"' in value:
            raise ReasonError(f"Property {name} cannot contain double quotes: {value}")

        super().__setattr__(name, value)

    def __str__(self) -> str:
        """String representation of the instance, including all attributes.

        Examples:
            * Example return value when the task ID is not set::

                Given reason - username@hostname

            * Example return value when the task ID is set::

                Given reason - username@hostname - T12345

        """
        parts = [self._reason, self.owner]
        if self._task_id is not None:
            parts.append(self._task_id)

        return " - ".join(parts)

    @property
    def reason(self) -> str:
        """Get the reason given to justify the administrative action."""
        return self._reason

    @property
    def owner(self) -> str:
        """Get the owner of the currently running code.

        Examples:
            Example return value::

                username@hostname

        """
        return f"{self._username}@{self._hostname}"

    @property
    def hostname(self) -> str:
        """Get  the hostname on which the code is running."""
        return self._hostname

    @property
    def task_id(self) -> Optional[str]:
        """Get the task ID to mention in the reason, or :py:data:`None` if none was given."""
        return self._task_id

    def quoted(self) -> str:
        """Get a double quoted string representation of the instance, including all attributes.

        Examples:
            * Example return value when the task ID is not set::

                "Given reason - username@hostname"

            * Example return value when the task ID is set::

                "Given reason - username@hostname - T12345"

        """
        return f'"{self}"'



"""Ganeti module."""

import logging
from dataclasses import dataclass
from typing import Optional, Union

from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException
from wmflib.requests import http_session

from spicerack.constants import PUPPET_CA_PATH
from spicerack.exceptions import SpicerackError
from spicerack.netbox import Netbox
from spicerack.remote import Remote, RemoteHosts

logger = logging.getLogger(__name__)

RAPI_URL_FORMAT: str = "https://{cluster}:5080"
"""The template string to construct the Ganeti RAPI URL."""
INSTANCE_LINKS: tuple[str, ...] = ("public", "private", "analytics")
"""The list of possible instance link types."""


class GanetiError(SpicerackError):
    """Raised on errors from Ganeti operations."""


@dataclass(frozen=True)
class GanetiCluster:
    """Represents a Ganeti cluster with all the related attributes.

    Arguments:
        name: the Ganeti cluster short name, equivalent to the Netbox cluster group name.
        fqdn: the FQDN of the Ganeti cluster VIP.
        rapi: the Ganeti RAPI endpoint URL to connect to.

    """

    name: str
    fqdn: str
    rapi: str


@dataclass(frozen=True)
class GanetiGroup:
    """Represents a Ganeti group with all the related attributes.

    Arguments:
        name: the Ganeti group name, equivalent to the Netbox cluster name.
        site: the Datacenter of the Ganeti group short name, equivalent of the Netbox site slug.
        cluster: the Ganeti cluster the group belongs to.

    """

    name: str
    site: str
    cluster: GanetiCluster


class GanetiRAPI:
    """Class which wraps the read-only Ganeti RAPI."""

    def __init__(self, cluster_url: str, username: str, password: str, timeout: int, ca_path: str):
        """Initialize the instance.

        Arguments:
            cluster_url: the URL of the RAPI endpoint.
            username: the RAPI user name
            password: the RAPI user's password
            timeout: the timeout in seconds for each request
            ca_path: the path to the signing certificate authority

        """
        self._url = cluster_url
        self._http_session = http_session(".".join((self.__module__, self.__class__.__name__)), timeout=timeout)
        self._http_session.auth = HTTPBasicAuth(username, password)
        self._http_session.verify = ca_path

    def _api_get_request(self, *targets: str) -> dict:
        """Perform a RAPI request and return its JSON decoded response.

        Arguments:
           *targets: The path components of the request (minus the /2/ part of the path).

        Raises:
           spicerack.ganeti.GanetiError: on non-200 responses.

        """
        full_url = "/".join([self._url, "2"] + list(targets))
        try:
            result = self._http_session.get(full_url)
        except RequestException as ex:
            raise GanetiError("Error while performing request to RAPI") from ex

        if result.status_code != 200:
            raise GanetiError(f"Non-200 from API: {result.status_code}: {result.text}")

        return result.json()

    @property
    def info(self) -> dict:
        """Return the complete cluster information.

        Raises:
            spicerack.ganeti.GanetiError: on API errors.

        """
        return self._api_get_request("info")

    @property
    def master(self) -> Optional[str]:
        """Return the internal name for the current ganeti master node or none if the data is missing.

        Raises:
            spicerack.ganeti.GanetiError: on API errors.

        """
        return self.info.get("master")

    def fetch_instance(self, fqdn: str) -> dict:
        """Return full information about an instance.

        Arguments:
           fqdn: the FQDN of the instance in question.

        Raises:
           spicerack.ganeti.GanetiError: on API errors.

        """
        return self._api_get_request("instances", fqdn)

    def fetch_instance_mac(self, fqdn: str) -> str:
        """Convenience method to return the 0th adapter's MAC address for an instance.

        Note that we don't allow the creation of instances with more than one MAC address at this time.

        Arguments:
           fqdn: the FQDN of the instance in question.

        Raises:
            spicerack.ganeti.GanetiError: on API errors.

        """
        instance_info = self.fetch_instance(fqdn)
        if "nic.macs" not in instance_info or not instance_info["nic.macs"]:
            raise GanetiError("Can't find any MACs for instance")

        return instance_info["nic.macs"][0]

    def groups(self, bulk: bool = False) -> dict:
        """Get a list of Cluster groups.

        Arguments:
            bulk: if true set bulk=1 to return detailed information.
                see https://docs.ganeti.org/docs/ganeti/2.9/html/rapi.html#bulk

        """
        target = "groups?bulk=1" if bulk else "groups"
        return self._api_get_request(target)

    def nodes(self, bulk: bool = False) -> dict:
        """Get a list of Cluster nodes.

        Arguments:
            bulk: if true set bulk=1 to return detailed information.
                see https://docs.ganeti.org/docs/ganeti/2.9/html/rapi.html#bulk

        """
        target = "nodes?bulk=1" if bulk else "nodes"
        return self._api_get_request(target)


class GntInstance:
    """Class that wraps gnt-instance command execution on a Ganeti cluster master host."""

    def __init__(self, master: RemoteHosts, cluster: str, instance: str):
        """Initialize the instance.

        Arguments:
            master: the Ganeti cluster master remote instance.
            cluster: the Ganeti cluster name.
            instance: the FQDN of the Ganeti VM instance to act upon.

        """
        self._master = master
        self._cluster = cluster
        self._instance = instance

    @property
    def cluster(self) -> str:
        """Getter for the Ganeti cluster name the instance belongs to."""
        return self._cluster

    def shutdown(self, *, timeout: int = 2) -> None:
        """Shutdown the Ganeti VM instance.

        Arguments:
            timeout: time in minutes to wait for a clean shutdown before pulling the plug.

        """
        logger.info("Shutting down VM %s in cluster %s", self._instance, self._cluster)
        self._master.run_sync(f"gnt-instance shutdown --force --timeout={timeout} {self._instance}")

    def startup(self) -> None:
        """Start the Ganeti VM instance."""
        logger.info("Starting VM %s in cluster %s", self._instance, self._cluster)
        self._master.run_sync(f"gnt-instance startup --force {self._instance}")

    def set_boot_media(self, boot: str) -> None:
        """Set the boot media of the Ganeti VM to the given media.

        Arguments:
            boot: the boot media to use. Use `disk` to boot from disk and `network` to boot from PXE.

        """
        self._master.run_sync(f"gnt-instance modify --hypervisor-parameters=boot_order={boot} {self._instance}")
        logger.info("Set boot media to %s for VM %s in cluster %s", boot, self._instance, self._cluster)

    def remove(self, *, shutdown_timeout: int = 2) -> None:
        """Shutdown and remove the VM instance from the Ganeti cluster, including its disks.

        Arguments:
            shutdown_timeout: time in minutes to wait for a clean shutdown before pulling the plug.

        Note:
            This action requires few minutes, inform the user about the waiting time when using this method.

        """
        logger.info(
            "Removing VM %s in cluster %s. This may take a few minutes.",
            self._instance,
            self._cluster,
        )
        self._master.run_sync(f"gnt-instance remove --shutdown-timeout={shutdown_timeout} --force {self._instance}")

    def add(self, *, group: str, vcpus: int, memory: Union[int, float], disk: int, link: str) -> None:
        """Create the VM for the instance in the Ganeti cluster with the specified characteristic.

        Arguments:
            group: the Ganeti group that matches the Datacenter physical row or rack where to allocate the instance.
            vcpus: the number of virtual CPUs to assign to the instance.
            memory: the amount of RAM to assign to the instance in gigabytes.
            disk: the amount of disk to assign to the instance in gigabytes.
            link: the type of network link to use, one of :py:const:`spicerack.ganeti.INSTANCE_LINKS`.

        Raises:
            spicerack.ganeti.GanetiError: on parameter validation error.

        Note:
            This action requires few minutes, inform the user about the waiting time when using this method.

        """
        if link not in INSTANCE_LINKS:
            raise GanetiError(f"Invalid link '{link}', expected one of: {INSTANCE_LINKS}")

        local_vars = locals()
        for var_label in ("vcpus", "memory", "disk"):
            if local_vars[var_label] <= 0:
                raise GanetiError(
                    f"Invalid value '{local_vars[var_label]}' for {var_label}, expected positive integer."
                )

        memory_mb = int(memory * 1024)
        command = (
            "gnt-instance add"
            " -t drbd"
            " -I hail"
            f" --net 0:link={link}"
            " --hypervisor-parameters=kvm:boot_order=network"
            " -o debootstrap+default"
            " --no-install"
            " --no-wait-for-sync"
            f" -g {group}"
            f" -B vcpus={vcpus},memory={memory_mb}m"
            f" --disk 0:size={disk}g"
            f" {self._instance}"
        )

        logger.info(
            (
                "Creating VM %s in cluster %s with group=%s vcpus=%d memory=%dGB disk=%dGB link=%s. "
                "This may take a few minutes."
            ),
            self._instance,
            self._cluster,
            group,
            vcpus,
            memory,
            disk,
            link,
        )

        results = self._master.run_sync(command, print_output=True)
        for _, output in results:
            logger.debug(output.message().decode())


class Ganeti:
    """Class which wraps all Ganeti clusters operations."""

    def __init__(self, username: str, password: str, timeout: int, remote: Remote, netbox: Netbox):
        """Initialize the instance.

        Arguments:
            username: The RAPI username to use.
            password: The RAPI password to use.
            timeout: The timeout in seconds for each request to the API.
            remote: the remote instance to connect to Ganeti hosts.
            netbox: the Netbox instance to gather data from the source of truth.

        """
        self._username = username
        self._password = password
        self._timeout = timeout
        self._remote = remote
        self._netbox = netbox

    def get_cluster(self, name: str) -> GanetiCluster:
        """Get a GanetiCluster instance for the given cluster name.

        Arguments:
            name: the name of the Ganeti cluster, equivalent to the cluster group in Netbox.

        Raises:
            spicerack.ganeti.GanetiError: if unable to find the cluster endpoint.

        """
        cluster_group = self._netbox.api.virtualization.cluster_groups.get(name=name)
        if cluster_group is None:
            raise GanetiError(f"Unable to find virtualization cluster group {name} on Netbox.")

        address_field = cluster_group.custom_fields.get("ip_address")
        if address_field is None:
            raise GanetiError(f"Virtualization cluster group {name} has no IP address.")

        address = address_field.get("address")
        if not address:  # Covers also the case it's an empty string
            raise GanetiError(f"Virtualization cluster group {name} IP address has no address.")

        ip_address = self._netbox.api.ipam.ip_addresses.get(address=address)
        if ip_address is None:
            raise GanetiError(f"Unable to find the IP address for the virtualization cluster group {name}.")

        if not ip_address.dns_name:
            raise GanetiError(f"Virtualization cluster group {name}'s IP address {address} has no DNS name.")

        return GanetiCluster(
            name=name, fqdn=ip_address.dns_name, rapi=RAPI_URL_FORMAT.format(cluster=ip_address.dns_name)
        )

    def get_group(self, name: str, *, cluster: str) -> GanetiGroup:
        """Get a GanetiGroup instance for the given group name.

        Arguments:
            name: the name of the Ganeti group, equivalent to the cluster in Netbox.
            cluster: the name of the Ganeti cluster where to look for the group, equivalent to the cluster group in
                Netbox.

        Raises:
            spicerack.ganeti.GanetiError: if unable to find the group.

        """
        cluster_obj = self.get_cluster(cluster)
        group = self._netbox.api.virtualization.clusters.get(name=name, group=cluster)
        if group is None:
            raise GanetiError(f"Unable to find virtualization cluster {name} in cluster group {cluster} on Netbox.")

        return GanetiGroup(name=name, site=group.site.slug, cluster=cluster_obj)

    def rapi(self, cluster: str) -> GanetiRAPI:
        """Return a RAPI object for a particular cluster.

        Arguments:
            cluster: the name of the cluster group in Netbox for this Ganeti cluster.

        Raises:
            spicerack.ganeti.GanetiError: if unable to find the cluster endpoint.

        """
        return GanetiRAPI(self.get_cluster(cluster).rapi, self._username, self._password, self._timeout, PUPPET_CA_PATH)

    def instance(self, instance: str, *, cluster: str = "") -> GntInstance:
        """Return an instance of GntInstance to perform RW operation on the given Ganeti VM instance.

        Arguments:
            instance: the FQDN of the Ganeti VM instance to act upon.
            cluster: the name of the Ganeti cluster to which the instance belongs, or will belong in case of a new
                instance to be created. If not provided it will be auto-detected for existing instances.

        """
        if not cluster:
            vm = self._netbox.api.virtualization.virtual_machines.get(name=instance.split(".")[0])
            if not vm:
                raise GanetiError(
                    f"Ganeti Virtual Machine {instance} does not exist on Netbox and no manual cluster was provided"
                )

            cluster = vm.cluster.group.name

        master = self.rapi(cluster).master
        if master is None:
            raise GanetiError(f"Master for cluster {cluster} is None")

        return GntInstance(self._remote.query(master), cluster, instance)



"""Decorators module."""
import inspect
import logging
from collections.abc import Callable
from typing import Any

from wmflib.decorators import RetryParams, ensure_wrap
from wmflib.decorators import retry as wmflib_retry

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


def get_effective_tries(params: RetryParams, func: Callable, args: tuple, kwargs: dict) -> None:
    """Reduce the number of tries to use in the @retry decorator to one when the DRY-RUN mode is detected.

    This is a callback function for the wmflib.decorators.retry decorator.
    The arguments are according to :py:func:`wmflib.decorators.retry` for the ``dynamic_params_callbacks`` argument.

    Arguments:
        params: the decorator original parameters.
        func: the decorated callable.
        args: the decorated callable positional arguments as tuple.
        kwargs: the decorated callable keyword arguments as dictionary.

    """
    reduce_tries = False
    qualname = getattr(func, "__qualname__", "")
    # Detect if func is an instance method that has self ensuring that the name of the class of the first parameter
    # (self) matches the class name extracted from the function's qualname.
    has_self = (
        "." in qualname and args and qualname.rsplit(".", 1)[0] == getattr(getattr(args[0], "__class__"), "__name__")
    )

    # When the decorated object is an instance method
    if has_self and (
        getattr(args[0], "_dry_run", False)  # Has self._dry_run
        or getattr(getattr(args[0], "_remote_hosts", False), "_dry_run", False)  # Has self._remote_hosts
    ):
        reduce_tries = True

    # When the decorated object is a function or method
    if "dry_run" in kwargs:  # Has an explicit dry_run parameter that was set in by the caller, use this one
        if kwargs["dry_run"] is True:
            reduce_tries = True
    else:  # Check if has a dry_run parameter with a default value
        signature_params = inspect.signature(func).parameters
        if "dry_run" in signature_params and signature_params["dry_run"].default is True:
            reduce_tries = True

    if reduce_tries:
        logger.warning("Reduce tries from %d to 1 in DRY-RUN mode", params.tries)
        params.tries = 1


def set_tries(params: RetryParams, _func: Callable, _params: tuple, kwargs: dict) -> None:
    """Simple function to allow adapting the number of retries.

    Arguments:
        params: the decorator original parameters.
        _func: the decorated callable. Unused.
        _args: the decorated callable positional arguments as tuple. Unused.
        kwargs: the decorated callable keyword arguments as dictionary.

    """
    override_default_retries = kwargs.get("tries", 0)
    if override_default_retries:
        params.tries = override_default_retries


@ensure_wrap
def retry(*args: Any, **kwargs: Any) -> Callable:
    """Decorator to retry a function or method if it raises certain exceptions with customizable backoff.

    A customized version of :py:func:`wmflib.decorators.retry` specific to Spicerack:

      * If no exceptions to catch are specified use :py:class:`spicerack.exceptions.SpicerackError`.
      * Always append to the ``dynamic_params_callbacks`` parameter the
        :py:func:`spicerack.decorators.get_effective_tries` function to force the tries parameter to 1 in
        DRY-RUN mode. Appending means that this callback will always be called for last, and eventually override any
        other modification of the tries parameter by other callbacks to 1 when in DRY-RUN mode.

    For the arguments see :py:func:`wmflib.decorators.retry`.

    Returns:
        The decorated function.

    """
    kwargs["dynamic_params_callbacks"] = tuple(list(kwargs.get("dynamic_params_callbacks", [])) + [get_effective_tries])
    kwargs["exceptions"] = kwargs.get("exceptions", (SpicerackError,))

    return wmflib_retry(*args[1:], **kwargs)(args[0])



"""Ganeti Module test."""

import json
from unittest import mock

import pytest
import requests
from ClusterShell.MsgTree import MsgTreeElem
from cumin import nodeset

from spicerack import ganeti
from spicerack.netbox import Netbox
from spicerack.remote import Remote
from spicerack.tests import get_fixture_path

TEST_CLUSTERS = {
    "sitea": "ganeti01.svc.sitea.example.com",
    "siteb01": "ganeti01.svc.siteb.example.com",
    "siteb02": "ganeti02.svc.siteb.example.com",
    "sitea_test": "ganeti-test01.svc.sitea.example.com",
}


class TestGaneti:  # pylint: disable=too-many-instance-attributes
    """Ganeti tests class."""

    def setup_method(self):
        """Setup test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.remote = mock.MagicMock(spec_set=Remote)
        self.netbox = mock.MagicMock(spec_set=Netbox)
        self.ganeti = ganeti.Ganeti(
            username="user", password="pass", timeout=10, remote=self.remote, netbox=self.netbox
        )

        self.cluster = "sitea"
        self.instance = "test.example.com"
        self.cluster_base_urls = {
            cluster: ganeti.RAPI_URL_FORMAT.format(cluster=url) + "/2" for cluster, url in TEST_CLUSTERS.items()
        }
        self.base_url = self.cluster_base_urls[self.cluster]
        self.instance_url = f"{self.base_url}/instances/{self.instance}"

        # Populate mock with default values
        self.netbox.api.virtualization.cluster_groups.get.return_value.custom_fields = {
            "ip_address": {"address": "10.0.0.1/24"}
        }
        self.netbox.api.ipam.ip_addresses.get.return_value.dns_name = TEST_CLUSTERS[self.cluster]
        self.netbox.api.virtualization.virtual_machines.get.return_value.cluster.group.name = self.cluster
        self.netbox.api.virtualization.clusters.get.return_value.site.slug = self.cluster

        # load test fixtures
        with open(get_fixture_path("ganeti", "info.json"), encoding="utf-8") as info_json:
            self.info = info_json.read()
        with open(get_fixture_path("ganeti", "404.json"), encoding="utf-8") as fourohfour_json:
            self.fourohfour = fourohfour_json.read()
        with open(get_fixture_path("ganeti", "instance.json"), encoding="utf-8") as instance_json:
            self.instance_info = instance_json.read()
        with open(get_fixture_path("ganeti", "bogus.json"), encoding="utf-8") as bogus_json:
            self.bogus_data = bogus_json.read()
        with open(get_fixture_path("ganeti", "groups.json"), encoding="utf-8") as groups_json:
            self.groups_data = groups_json.read()
        with open(get_fixture_path("ganeti", "groups_bulk.json"), encoding="utf-8") as groups_bulk_json:
            self.groups_bulk_data = groups_bulk_json.read()
        with open(get_fixture_path("ganeti", "nodes.json"), encoding="utf-8") as nodes_json:
            self.nodes_data = nodes_json.read()
        with open(get_fixture_path("ganeti", "nodes_bulk.json"), encoding="utf-8") as nodes_bulk_json:
            self.nodes_bulk_data = nodes_bulk_json.read()

    def _set_requests_mock_for_instance(self, requests_mock, missing_active=False):
        """Set request mock to be 404 on all other clusters."""
        for cluster in TEST_CLUSTERS:
            if cluster == self.cluster and not missing_active:
                requests_mock.get(self.instance_url, text=self.instance_info)
            else:
                requests_mock.get(
                    f"{self.cluster_base_urls[cluster]}/instances/{self.instance}",
                    text=self.fourohfour,
                    status_code=requests.codes["not_found"],
                )

    @pytest.mark.parametrize("cluster", TEST_CLUSTERS)
    def test_rapi_clusters_ok(self, cluster):
        """Ganeti.rapi() should return a GanetiRAPI object."""
        self.netbox.api.virtualization.cluster_groups.get.return_value.custom_fields = {
            "ip_address": {"address": "10.0.0.1/24"}
        }
        self.netbox.api.ipam.ip_addresses.get.return_value.dns_name = TEST_CLUSTERS[cluster]
        assert isinstance(self.ganeti.rapi(cluster), ganeti.GanetiRAPI)

    def test_rapi_clusters_missing_cluster(self):
        """If unable to find the cluster in Netbox it should raise a GanetiError."""
        self.netbox.api.virtualization.cluster_groups.get.return_value = None
        with pytest.raises(ganeti.GanetiError, match="Unable to find virtualization cluster group invalid on Netbox"):
            self.ganeti.rapi("invalid")

    def test_rapi_clusters_no_ip_address(self):
        """If the cluster has no IP address in Netbox it should raise a GanetiError."""
        self.netbox.api.virtualization.cluster_groups.get.return_value.custom_fields = {"ip_addresses": None}
        with pytest.raises(ganeti.GanetiError, match="Virtualization cluster group sitea has no IP address"):
            self.ganeti.rapi("sitea")

    @pytest.mark.parametrize(
        "address",
        (
            {"ip_address": {}},
            {"ip_address": {"address": None}},
            {"ip_address": {"address": ""}},
        ),
    )
    def test_rapi_clusters_no_address(self, address):
        """If the cluster IP address has no address in Netbox it should raise a GanetiError."""
        self.netbox.api.virtualization.cluster_groups.get.return_value.custom_fields = address
        with pytest.raises(ganeti.GanetiError, match="Virtualization cluster group sitea IP address has no address"):
            self.ganeti.rapi("sitea")

    def test_rapi_clusters_address_not_found(self):
        """If the IP address of the cluster group cannot be found in Netbox it should raise a GanetiError."""
        self.netbox.api.virtualization.cluster_groups.get.return_value.custom_fields = {
            "ip_address": {"address": "10.0.0.1/24"}
        }
        self.netbox.api.ipam.ip_addresses.get.return_value = None
        with pytest.raises(
            ganeti.GanetiError, match="Unable to find the IP address for the virtualization cluster group sitea"
        ):
            self.ganeti.rapi("sitea")

    @pytest.mark.parametrize("dns_name", (None, ""))
    def test_rapi_clusters_address_no_dns_name(self, dns_name):
        """If the IP address of the cluster group doesn't have a DNS name it should raise a GanetiError."""
        self.netbox.api.virtualization.cluster_groups.get.return_value.custom_fields = {
            "ip_address": {"address": "10.0.0.1/24"}
        }
        self.netbox.api.ipam.ip_addresses.get.return_value.dns_name = dns_name
        with pytest.raises(
            ganeti.GanetiError, match="Virtualization cluster group sitea's IP address 10.0.0.1/24 has no DNS name"
        ):
            self.ganeti.rapi("sitea")

    def test_rapi_master_ok(self, requests_mock):
        """The master property of a RAPI should be the hostname for the master of this cluster."""
        requests_mock.get(self.base_url + "/info", text=self.info)

        master = json.loads(self.info)["master"]
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master == master

    def test_rapi_master_missing(self, requests_mock):
        """If a master is not specified by the uptsream API, the value of master on a RAPI should be None."""
        requests_mock.get(self.base_url + "/info", text=self.bogus_data)
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master is None

    def test_rapi_instance_missing(self, requests_mock):
        """A RAPI object should raise a GanetiError if a requested host does not exist."""
        requests_mock.get(
            self.instance_url,
            text=self.fourohfour,
            status_code=requests.codes["not_found"],
        )
        rapi = self.ganeti.rapi(self.cluster)
        with pytest.raises(ganeti.GanetiError, match=r"Non-200 from API: 404:.*"):
            rapi.fetch_instance(self.instance)

    def test_rapi_instance_request_fail(self, requests_mock):
        """A RAPI object should raise a GanetiError if a request fails."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(ganeti.GanetiError, match="Error while performing request to RAPI"):
            rapi.fetch_instance(self.instance)

    def test_rapi_instance_invalid(self, requests_mock):
        """If no mac is present in host data, fetch_instance_mac should raise a GanetiError."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, text=self.bogus_data)
        with pytest.raises(ganeti.GanetiError, match=""):
            rapi.fetch_instance_mac(self.instance)

    def test_rapi_instance_valid(self, requests_mock):
        """The MAC returned by RAPI.fetch_instance_mac should match the data returned by the API."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, text=self.instance_info)
        mac = json.loads(self.instance_info)["nic.macs"][0]
        assert mac == rapi.fetch_instance_mac(self.instance)

    @pytest.mark.parametrize("bulk", (True, False))
    def test_rapi_nodes(self, requests_mock, bulk):
        """The master property of a RAPI should be the hostname for the master of this cluster."""
        if bulk:
            uri = "/nodes?bulk=1"
            data = self.nodes_bulk_data
        else:
            uri = "/nodes"
            data = self.nodes_data
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.base_url + uri, text=data)

        nodes = json.loads(data)
        assert rapi.nodes(bulk) == nodes

    @pytest.mark.parametrize("bulk", (True, False))
    def test_rapi_groups(self, requests_mock, bulk):
        """The master property of a RAPI should be the hostname for the master of this cluster."""
        if bulk:
            uri = "/groups?bulk=1"
            data = self.groups_bulk_data
        else:
            uri = "/groups"
            data = self.groups_data
        requests_mock.get(self.base_url + uri, text=data)

        groups = json.loads(data)
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.groups(bulk) == groups

    @pytest.mark.parametrize("cluster", ("", "sitea"))
    def test_instance_ok(self, cluster, requests_mock):
        """It should return an instance of GntInstance for a properly configured cluster."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance, cluster=cluster)
        assert isinstance(instance, ganeti.GntInstance)
        assert instance.cluster == "sitea"
        self.remote.query.assert_called_once_with("ganeti1.example.com")

    def test_instance_missing_vm_no_cluster(self):
        """It should raise a GanetiError if the VM does not exist in Netbox and no cluster is provided."""
        self.netbox.api.virtualization.virtual_machines.get.return_value = None
        with pytest.raises(
            ganeti.GanetiError, match="Ganeti Virtual Machine test.example.com does not exist on Netbox and"
        ):
            self.ganeti.instance(self.instance)

        self.remote.query.assert_not_called()

    def test_instance_missing_master(self, requests_mock):
        """It should raise a GanetiError exception if unable to determin the instance's master to manage it."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.bogus_data)
        with pytest.raises(ganeti.GanetiError, match="Master for cluster sitea is None"):
            self.ganeti.instance(self.instance)

    def test_instance_startup(self, requests_mock):
        """It should issue the startup command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.startup()
        self.remote.query.return_value.run_sync.assert_called_once_with("gnt-instance startup --force test.example.com")

    def test_instance_set_boot_media(self, requests_mock):
        """It should set the boot media to the one provided."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.set_boot_media("disk")
        self.remote.query.return_value.run_sync.assert_called_once_with(
            "gnt-instance modify --hypervisor-parameters=boot_order=disk test.example.com"
        )

    @pytest.mark.parametrize("kwargs", ({}, {"timeout": 0}))
    def test_instance_shutdown(self, requests_mock, kwargs):
        """It should issue the shutdown command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.shutdown(**kwargs)
        timeout = kwargs["timeout"] if "timeout" in kwargs else 2
        self.remote.query.return_value.run_sync.assert_called_once_with(
            f"gnt-instance shutdown --force --timeout={timeout} test.example.com"
        )

    @pytest.mark.parametrize("kwargs", ({}, {"shutdown_timeout": 0}))
    def test_instance_remove(self, requests_mock, kwargs):
        """It should issue the remove command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.remove(**kwargs)
        timeout = kwargs["shutdown_timeout"] if "shutdown_timeout" in kwargs else 2
        self.remote.query.return_value.run_sync.assert_called_once_with(
            f"gnt-instance remove --shutdown-timeout={timeout} --force test.example.com"
        )

    def test_instance_add_ok(self, requests_mock):
        """It should issue the remove command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        results = [
            (
                nodeset("ganeti-master.example.com"),
                MsgTreeElem(b"creation logs", parent=MsgTreeElem()),
            )
        ]
        self.remote.query.return_value.run_sync.return_value = iter(results)

        instance.add(group="row_A", vcpus=2, memory=3.1, disk=4, link="private")

        self.remote.query.return_value.run_sync.assert_called_once_with(
            "gnt-instance add -t drbd -I hail --net 0:link=private --hypervisor-parameters=kvm:boot_order=network "
            "-o debootstrap+default --no-install --no-wait-for-sync -g row_A -B vcpus=2,memory=3174m --disk 0:size=4g "
            "test.example.com",
            print_output=True,
        )

    @pytest.mark.parametrize(
        "kwargs, exc_message",
        (
            (
                {"group": "row_A", "vcpus": 1, "memory": 1, "disk": 1, "link": "invalid"},
                r"Invalid link 'invalid', expected one of: \('public', 'private', 'analytics'\)",
            ),
            (
                {"group": "row_A", "vcpus": -1, "memory": 1, "disk": 1, "link": "private"},
                r"Invalid value '-1' for vcpus, expected positive integer.",
            ),
            (
                {"group": "row_A", "vcpus": 1, "memory": -1, "disk": 1, "link": "private"},
                r"Invalid value '-1' for memory, expected positive integer.",
            ),
            (
                {"group": "row_A", "vcpus": 1, "memory": 1, "disk": -1, "link": "private"},
                r"Invalid value '-1' for disk, expected positive integer.",
            ),
        ),
    )
    def test_instance_add_fail(self, requests_mock, kwargs, exc_message):
        """It should raise GanetiError on invalid parameters."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        with pytest.raises(ganeti.GanetiError, match=exc_message):
            instance.add(**kwargs)

        assert not self.remote.query.return_value.run_sync.called

    def test_get_group_ok(self):
        """It should return a GanetiGroup instance."""
        group = self.ganeti.get_group("group1", cluster=self.cluster)
        assert isinstance(group, ganeti.GanetiGroup)
        assert isinstance(group.cluster, ganeti.GanetiCluster)
        assert group.name == "group1"
        assert group.site == self.cluster
        assert group.cluster.name == self.cluster
        assert group.cluster.fqdn == TEST_CLUSTERS[self.cluster]
        assert group.cluster.rapi == f"https://{TEST_CLUSTERS[self.cluster]}:5080"

    def test_get_group_fail(self):
        """It should raise a GanetiError if the group doesn't exists."""
        self.netbox.api.virtualization.clusters.get.return_value = None
        with pytest.raises(
            ganeti.GanetiError, match="Unable to find virtualization cluster group1 in cluster group sitea on Netbox"
        ):
            self.ganeti.get_group("group1", cluster=self.cluster)



"""Group3 argument_parser() raise SystemExit."""
import argparse

__title__ = __doc__


def argument_parser():
    """As required by spicerack._cookbook."""
    parser = argparse.ArgumentParser("group3.argument_parser_raise_system_exit")
    parser.add_argument("required")
    return parser


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""
